"""
Bohrium Loop
============
"""

import sys
import copy
from functools import reduce
import numpy_force as numpy
from . import bhary


class Iterator(object):
    """Iterator used for sliding views within loops.

    Notes
    -----
    Supports addition, subtraction and multiplication.
    """

    def __init__(self, max_iter, value, step_delay=1, reset=None):
        """The initial state of the iterator.

        Parameters
        ----------
        max_iter : int
            The maximum amount of iterations the loop can go on for
        value : int
            The beginning offset of the iterator
        step_delay : int
            The amount of iterations needed before a change is made
        reset : int
            The amount of iterations before the changes are reset

        Notes
        -----
        step : int
            The amount the offset is slided when performing a step

        The step can be changed by using multiplication
        """

        self.max_iter = max_iter
        self.offset = value
        self.step_delay = step_delay
        self.reset = reset

        self.step = 1

    def __add__(self, other):
        new_it = copy.copy(self)
        new_it.offset += other
        return new_it

    def __radd__(self, other):
        new_it = copy.copy(self)
        new_it.offset += other
        return new_it

    def __sub__(self, other):
        new_it = copy.copy(self)
        new_it.offset -= other
        return new_it

    def __rsub__(self, other):
        new_it = copy.copy(self)
        new_it.step *= -1
        new_it.offset = other - new_it.offset
        return new_it

    def __mul__(self, other):
        new_it = copy.copy(self)
        new_it.offset *= other
        new_it.step *= other
        return new_it

    def __rmul__(self, other):
        new_it = copy.copy(self)
        new_it.offset *= other
        new_it.step *= other
        return new_it

    def __neg__(self):
        new_it = copy.copy(self)
        new_it.step *= -1
        new_it.offset *= -1
        return new_it


class IteratorOutOfBounds(Exception):
    """Exception thrown when a view goes out of bounds after the maximum
       iterations."""

    def __init__(self, dim, shape, first_index, last_index):
        error_msg = \
            "\n    Iterator out of bounds:\n" \
            "     Dimension %d has length %d, iterator starts from %d and goes to %d." \
            % (dim, shape, first_index, last_index)
        super(IteratorOutOfBounds, self).__init__(error_msg)


class IteratorIllegalDepth(Exception):
    """Exception thrown when a view consists of a mix of iterator depths."""

    def __init__(self):
        error_msg = \
            "\n    Illegal mix of iterators with different depth:\n" \
            "     A view cannot use iterators from different depths of the grid in the same dimension."
        super(IteratorIllegalDepth, self).__init__(error_msg)


class IteratorIllegalBroadcast(Exception):
    """Exception thrown when a view consists of a mix of iterator depths."""

    def __init__(self, dim, a_shape, a_shape_change,
                 bcast_shape, bcast_shape_change):
        error_msg = \
            "Broadcast with dynamic shape:\n" \
            "    View with shape " + str(a_shape) + \
            " changes shape by " + str(a_shape_change) + \
            " in dimension " + str(dim) + \
            ".\n    It is differet from the view it is broadcasted from with shape " + \
            str(bcast_shape) + \
            " which changes shape by " + str(bcast_shape_change) + ".\n"
        super(IteratorIllegalBroadcast, self).__init__(error_msg)


class DynamicViewInfo(object):
    """Object for storing information about dynamic changes to the view"""

    def __init__(self, dynamic_changes, shape, stride, resets={}):
        """The initial state of the dynamic view information.

        Parameters
        ----------
        dynamic_changes : {int : [(int, int, int, int, int)]}
            A dictionary of lists of tuples corresponding to the dynamic
            changes the view.
            The tuple is (slide, shape_change, step_delay, shape, stride)
            and is explained further in the parameters to `add_dynamic_change`.
        shape : int tuple (variable length)
            The shape of the view that the dynamic view is based upon.
            Used when using negative indices to wrap around.
        stride : int tuple (variable length)
            The stride of the view that the dynamic view is based upon.
            Used for making slides to the offset.
        resets : {int : int}
            A dictionary of ints. The key corresponds to the dimension and
            the value to iterations before the changes are reset.
            Used for nested loops.
        """
        self.shape = shape
        self.stride = stride
        self.dynamic_changes = dynamic_changes
        self.resets = resets

    def add_dynamic_change(self, dim, slide, shape_change, step_delay, shape=None, stride=None):
        """Add dynamic changes to the dynamic changes information of the view.

        Parameters
        ----------
        dim : int
            The relevant dimension
        slide : int
            The change to the offset in the given dimension
            (can be both positive and negative)
        shape_change : int
            The amount the shape changes in the given dimension
            (can also be both positive and negative)
        step_delay : int
            If the change is based on an iterator in a grid, it is
            the changes can be delayed until the inner iterators
            have been updated `step_delay` times.
        shape : int
            The shape that the view can slide within. If not given,
            self.shape[dim] is used instead
        stride : int
            The stride that the view can slide within. If not given,
            self.stride[dim] is used instead
        """
        if not shape:
            shape = self.shape[dim]
        if not stride:
            stride = self.stride[dim]

        if self.has_changes_in_dim(dim):
            self.dynamic_changes[dim].append((slide, shape_change, step_delay, shape, stride))
        else:
            self.dynamic_changes[dim] = [(slide, shape_change, step_delay, shape, stride)]

    def has_changes_in_dim(self, dim):
        """Check whether there are any dynamic changes in the given dimension.

        Parameters
        ----------
        dim : int
            The relevant dimension
        """
        return dim in self.dynamic_changes

    def dims_with_changes(self):
        """Returns a list of all dimensions with dynamic changes."""
        return self.dynamic_changes.keys()

    def changes_in_dim(self, dim):
        """Returns a list of all dynamic changes in a dimension.
        If the dimension does not contain any dynamic changes,
        an empty list is returned.

        Parameters
        ----------
        dim : int
            The relevant dimension
        """
        changes = []
        if self.has_changes_in_dim(dim):
            changes += self.dynamic_changes[dim]
        return changes

    def index_into(self, dvi):
        """Modifies the dynamic change such that is reflects
        being indexed into another view with dynamic changes.

        Parameters
        ----------
        dim : DynamicViewInfo
            The infomation about dynamic changes within the
            view which is indexed into
        """
        for dim in dvi.dims_with_changes():
            if (dim in self.resets) or (dim in dvi.resets):
                if not (dim in dvi.resets) or \
                        not (dim in self.resets) or \
                        self.resets[dim] != dvi.resets[dim] or \
                        self.changes_in_dim(dim) != dvi.changes_in_dim(dim):
                    raise IteratorIllegalDepth()
            for change in dvi.changes_in_dim(dim):
                self.add_dynamic_change(dim, *change)

    def dim_shape_change(self, dim):
        """Returns the summed shape change in the given dimension.

        Parameters
        ----------
        dim : int
            The relevant dimension
        """
        shape_change_sum = 0
        if self.has_changes_in_dim(dim):
            for (_, shape_change, _, _, _) in self.dynamic_changes[dim]:
                shape_change_sum += shape_change
        return shape_change_sum

    def get_shape_changes(self):
        """Returns a dictionary of all changes to the shape.
        The dimension is the key and the shape change in the
        dimension is the value.
        """
        shape_changes = {}
        for dim in self.dynamic_changes.keys():
            shape_changes[dim] = 0
            for (_, shape_change, _, _, _) in self.dynamic_changes[dim]:
                shape_changes[dim] += shape_change
        return shape_changes

    def has_changes(self):
        """Returns whether the object contains any dynamic changes."""
        return self.dynamic_changes != {}


def inherit_dynamic_changes(a, sliced):
    """Creates a view into another view which has dynamic changes.
    The new view inherits the dynamic changes."""
    # Temporary store the dynamic changes
    dvi = a.bhc_dynamic_view_info

    # Perform slicing (removes the dynamic changes to avoid infinite recursion)
    a.bhc_dynamic_view_info = None
    b = a[sliced]

    # Inherit the dynamic changes (and restore dynamic changes to a)
    b.bhc_dynamic_view_info = dvi
    a.bhc_dynamic_view_info = dvi
    return b


def get_iterator(max_iter, val, step_delay=1):
    """Returns an iterator with a given starting value. An iterator behaves like
       an integer, but is used to slide view between loop iterations.

    Parameters
    ----------
    max_iter : int
        The maximum amount of iterations of the loop. Used for checking
        boundaries.
    val : int
        The initial value of the iterator.

    Notes
    -----
    `get_iterator` can only be used within a bohrium loop function. Within the
    loop `max_iter` is set by a lambda function. This is also the case in the
    example.

    Examples
    --------
    >>> def kernel(a):
    ...     i = get_iterator(1)
    ...     a[i] *= a[i-1]
    >>> a = bh.arange(1,6)
    >>> bh.do_while(kernel, 4, a)
    array([1, 2, 6, 24, 120])"""

    it = Iterator(max_iter, val, step_delay)
    return it


def get_grid(max_iter, *args):
    """Returns n iterators in a grid, corresponding to nested loops.

    Parameters
    ----------
    args : pointer to two or more integers
        The first integer is the maximum iterations of the loop, used for checking
        boundaries. The rest are the shape of the grid.

    Notes
    -----
    `get_grid` can only be used within a bohrium loop function. Within the
    loop `max_iter` is set by a lambda function.
    There are no upper bound on the amount of grid values.

    Examples
    --------
    >>> def kernel(a):
    ...     i, j, k = get_grid(3,3,3)
    ...     a[i,j,k] += 1

    correspondes to

    >>> for i in range(3):
    ...     for j in range(3):
    ...         for k in range(3):
    ...             a[i,j,k] += 1"""

    # Remove maximum iterations and reverse the grid to
    # loop over the grid from inner to outer
    grid = args[::-1]

    # Tuple of resulting iterators
    iterators = ()

    # Beginning step delay is always 1
    step_delay = 1

    for dim, iterations in enumerate(grid):
        i = Iterator(max_iter, 0, step_delay, iterations)
        step_delay *= iterations
        iterators = (i,) + iterators
    return iterators


def has_iterator(*s):
    """Checks whether a (multidimensional) slice contains an iterator

    Parameters
    ----------
    s : pointer to an integer, iterator or a tuple of integers/iterators

    Notes
    -----
    Only called from __getitem__ and __setitem__ in bohrium arrays (see _bh.c)."""

    # Helper function for one-dimensional slices
    def check_simple_type(ss):
        if isinstance(ss, slice):
            # Checks whether the view changes shape during iterations
            return isinstance(ss.start, Iterator) or \
                   isinstance(ss.stop, Iterator)
        else:
            return isinstance(ss, Iterator)

    # Checking single or multidimensional slices for iterators
    if isinstance(s, tuple):
        for t in s:
            it = check_simple_type(t)
            if it:
                return it
        return False
    else:
        return check_simple_type(s)


def slide_from_view(a, sliced):
    def check_bounds(shape, dim, s):
        """Checks whether the view is within the bounds of the array,
        given the maximum number of iterations"""

        # If the dimension is reset, then only the range before the reset needs to be checked
        if s.reset and s.max_iter / s.step_delay >= s.reset:
            last_index = s.offset + (s.reset - 1) * s.step
        else:
            last_index = s.offset + s.step / s.step_delay * (s.max_iter - 1)

        # Check that the starting index and the last index is within bounds
        if -shape[dim] <= s.offset < shape[dim] and \
                -shape[dim] <= last_index < shape[dim]:
            return True
        else:
            raise IteratorOutOfBounds(dim, shape[dim], s.offset, last_index)

    def dynamic_shape_change(s):
        """Returns how the shape of a view changes between iterations
        based on a slice possibly containing iterators"""
        if isinstance(s.start, Iterator):
            start_step = s.start.step
        else:
            start_step = 0
        if isinstance(s.stop, Iterator):
            stop_step = s.stop.step
        else:
            stop_step = 0
        return stop_step - start_step

    # Make sure that the indices is within a tuple
    if not isinstance(sliced, tuple):
        sliced = (sliced,)

    # Checks whether the indices contains a slice
    has_slices = reduce((lambda x, y: x or y), [isinstance(s, slice) for s in sliced])

    # The new slices (does not contain iterators)
    new_slices = ()
    # The dynamic changes
    slides = []
    # The resets (used for resetting an iterator in nested loops)
    resets = {}

    for i, s in enumerate(sliced):
        if len(sliced) == 1 or has_iterator(s):
            # A slice with optional step size (eg. a[i:i+2] or a[i:i+2:2])
            if isinstance(s, slice):
                start_is_iterator = isinstance(s.start, Iterator)
                stop_is_iterator = isinstance(s.stop, Iterator)

                # Cannot contain iterators with different reset in same slice
                if start_is_iterator and stop_is_iterator \
                        and (s.start.step_delay != s.stop.step_delay or s.start.reset != s.stop.reset):
                    raise IteratorIllegalDepth()

                # Check whether the start/end iterator stays within the array
                if start_is_iterator:
                    check_bounds(a.shape, i, s.start)
                    start = s.start.offset
                    step = s.start.step
                    step_delay = s.start.step_delay
                    reset = s.start.reset
                else:
                    start = s.start
                    step = 0
                    step_delay = 1

                if stop_is_iterator:
                    if s.stop.offset > 0:
                        check_bounds(a.shape, i, s.stop - 1)
                    else:
                        check_bounds(a.shape, i, s.stop)
                    stop = s.stop.offset
                    reset = s.stop.reset
                else:
                    stop = s.stop

                # Store the new slice
                new_slices += (slice(start, stop, s.step),)
                slides.append((i, step, dynamic_shape_change(s), step_delay))

            # A single iterator (eg. a[i])
            else:
                # Check whether the iterator stays within the array
                check_bounds(a.shape, i, s)

                # If the indices does not contain a slice, the returned value must
                # be a view with shape 1 in each dimension (To avoid a flush)
                if not has_slices:
                    new_slices += (slice(s.offset, s.offset + 1),)
                else:
                    new_slices += (s.offset,)
                slides.append((i, s.step, 0, s.step_delay))
                reset = s.reset

            # Add information about dimension being reset
            if reset:
                resets[i] = reset
        else:
            # Does not contain an iterator, just pass it through
            new_slices += (s,)

    # Use the indices to create a new view
    b = a[new_slices]

    a_dvi = a.bhc_dynamic_view_info

    if a_dvi:
        b_dvi = DynamicViewInfo({}, a_dvi.shape, a.strides, resets)
    else:
        b_dvi = DynamicViewInfo({}, a.shape, a.strides, resets)

    for slide in slides:
        b_dvi.add_dynamic_change(*slide)

    # If the view, which is indexed into, contains dynamic changes,
    # pass them on to the new view

    if a_dvi:
        b_dvi.index_into(a_dvi)

    if b_dvi.has_changes():
        b.bhc_dynamic_view_info = b_dvi
    return b


def add_slide_info(a):
    """Checks whether a view is dynamic and adds the relevant
       information to the view structure within BXX if it is.

    Parameters
    ----------
    a : array view
        A view into an array
    """
    from . import _bh

    # Check whether the view is a dynamic view
    dvi = a.bhc_dynamic_view_info

    if dvi:
        # Set the relevant update conditions for the new view
        for dim in dvi.dynamic_changes.keys():
            for (slide, shape_change, step_delay, shape, stride) in dvi.dynamic_changes[dim]:
                try:
                    stride = int(stride / 8)
                    shape = shape
                except:
                    stride = 0
                    shape = 0

                # Add dynamic information to the view within the cxx bridge
                _bh.slide_view(a, dim, slide, shape_change, shape, stride, step_delay)
        # Add resets to the relevant dimensions within the cxx bridge (used for nested loops)
        for dim in dvi.resets.keys():
            _bh.add_reset(a, dim, dvi.resets[dim])


def do_while(func, niters, *args, **kwargs):
    """Repeatedly calls the `func` with the `*args` and `**kwargs` as argument.

    The `func` is called while `func` returns True or None and the maximum number
    of iterations, `niters`, hasn't been reached.

    Parameters
    ----------
    func : function
        The function to run in each iterations. `func` can take any argument and may return
        a boolean `bharray` with one element.
    niters: int or None
        Maximum number of iterations in the loop (number of times `func` is called). If None, there is no maximum.
    *args, **kwargs : list and dict
        The arguments to `func`

    Notes
    -----
    `func` can only use operations supported natively in Bohrium.

    Examples
    --------
    >>> def loop_body(a):
    ...     a += 1
    >>> a = bh.zeros(4)
    >>> bh.do_while(loop_body, 5, a)
    >>> a
    array([5, 5, 5, 5])

    >>> def loop_body(a):
    ...     a += 1
    ...     return bh.sum(a) < 10
    >>> a = bh.zeros(4)
    >>> bh.do_while(loop_body, None, a)
    >>> a
    array([3, 3, 3, 3])
    """

    from . import _bh
    _bh.flush()
    flush_count = _bh.flush_count()
    func.__globals__['get_iterator'] = lambda x=0: get_iterator(niters, x)
    func.__globals__['get_grid'] = lambda *argz: get_grid(*((niters,) + argz))
    cond = func(*args, **kwargs)
    if flush_count != _bh.flush_count():
        raise TypeError("Invalid `func`: the looped function contains operations not support "
                        "by Bohrium, contain branches, or is simply too big!")
    if niters is None:
        niters = sys.maxsize - 1
    if cond is None:
        _bh.flush_and_repeat(niters, None)
    else:
        if not bhary.check(cond):
            raise TypeError("Invalid `func`: `func` may only return Bohrium arrays or nothing at all")
        if cond.dtype.type is not numpy.bool_:
            raise TypeError("Invalid `func`: `func` returned array of wrong type `%s`. "
                            "It must be of type `bool`." % cond.dtype)
        if len(cond.shape) != 0 and len(cond) > 1:
            raise TypeError("Invalid `func`: `func` returned array of shape `%s`. "
                            "It must be a scalar or an array with one element." % cond.shape)
        if not bhary.is_base(cond):
            raise TypeError("Invalid `func`: `func` returns an array view. It must return a base array.")

        _bh.sync(cond)
        _bh.flush_and_repeat(niters, cond)



"""
LinAlg
~~~~~~

Common linear algebra functions

"""
import bohrium as np
import bohrium.blas as blas
import numpy_force.linalg as la
import numpy_force as numpy

# We import all of NumPy LinAlg and overwrite with the objects we implement ourself
from numpy_force.linalg import *

from . import bhary
from . import ufuncs
from . import array_create
from . import user_kernel
from ._util import dtype_equal
from .bhary import fix_biclass_wrapper


@fix_biclass_wrapper
def gauss(a):
    """
    Performe Gausian elimination on matrix a without pivoting
    """
    for c in range(1, a.shape[0]):
        a[c:, c - 1:] = a[c:, c - 1:] - (a[c:, c - 1] / a[c - 1, c - 1:c])[:, None] * a[c - 1, c - 1:]
        np.flush()
    a /= np.diagonal(a)[:, None]
    return a


@fix_biclass_wrapper
def lu(a):
    """
    Performe LU decomposition on the matrix a so A = L*U
    """
    u = a.copy()
    l = np.zeros_like(a)
    np.diagonal(l)[:] = 1.0
    for c in range(1, u.shape[0]):
        l[c:, c - 1] = (u[c:, c - 1] / u[c - 1, c - 1:c])
        u[c:, c - 1:] = u[c:, c - 1:] - l[c:, c - 1][:, None] * u[c - 1, c - 1:]
        np.flush()
    return (l, u)


@fix_biclass_wrapper
def solve(a, b):
    """
    Solve a linear matrix equation, or system of linear scalar equations
    using Gausian elimination.

    :param a: Coefficient matrix
    :type a:  array_like, shape (M, M)
    :param b: Ordinate or "dependent variable" values
    :type b:  array_like, shape (M,) or (M, N)

    :return:  Solution to the system a x = b
    :rtype:   ndarray, shape (M,) or (M, N) depending on b

    :raises: :py:exc:`LinAlgError` If `a` is singular or not square.

    **Examples:**
    Solve the system of equations ``3 * x0 + x1 = 9`` and ``x0 + 2 * x1 = 8``:

    >>> import bohrium as np
    >>> a = np.array([[3.,1.], [1.,2.]])
    >>> b = np.array([9.,8.])
    >>> x = np.linalg.solve(a, b)
    >>> x
    array([ 2.,  3.])

    Check that the solution is correct:

    >>> (np.dot(a, x) == b).all()
    True
    """
    if not (len(a.shape) == 2 and a.shape[0] == a.shape[1]):
        raise la.LinAlgError("a is not square")

    w = gauss(np.hstack((a, b[:, np.newaxis])))
    lc = w.shape[1] - 1
    x = w[:, lc].copy()
    for c in range(lc - 1, 0, -1):
        x[:c] -= w[:c, c] * x[c:c + 1]
        np.flush()
    return x


@fix_biclass_wrapper
def jacobi(a, b, tol=0.0005):
    """
    Solve a linear matrix equation, or system of linear scalar equations
    using the Jacobi Method.

    :param a: Coefficient matrix
    :type a:  array_like, shape (M, M)
    :param b: Ordinate or "dependent variable" values
    :type b:  array_like, shape (M,) or (M, N)

    :return:  Solution to the system a x = b
    :rtype:   ndarray, shape (M,) or (M, N) depending on b

    :raises: :py:exc:`LinAlgError` If `a` is singular or not square.

    **Examples:**
    Solve the system of equations ``3 * x0 + x1 = 9`` and ``x0 + 2 * x1 = 8``:

    >>> import bohrium as np
    >>> a = np.array([[3,1], [1,2]])
    >>> b = np.array([9,8])
    >>> x = np.linalg.jacobi(a, b)
    >>> x
    array([ 2.,  3.])

    Check that the solution is correct:

    >>> (np.dot(a, x) == b).all()
    True
    """
    x = np.ones_like(b)
    D = 1 / np.diag(a)
    R = np.diag(np.diag(a)) - a
    T = D[:, np.newaxis] * R
    C = D * b
    error = tol + 1
    while error > tol:
        xo = x
        x = np.add.reduce(T * x, -1) + C
        error = norm(x - xo) / norm(x)
    return x


@fix_biclass_wrapper
def matmul(a, b, no_blas=False):
    """
    Matrix multiplication of two 2-D arrays.

    Parameters
    ----------
    a : array_like
        First argument.
    b : array_like
        Second argument.

    Returns
    -------
    output : ndarray
        Returns the matrix multiplication of `a` and `b`.

    Raises
    ------
    ValueError
        If the last dimension of `a` is not the same size as
        the second-to-last dimension of `b`.

    See Also
    --------
    dot : Dot product of two arrays.

    Examples
    --------
    >>> np.matmul(np.array([[1,2],[3,4]]),np.array([[5,6],[7,8]]))
    array([[19, 22],
           [43, 50]])
    """
    if not dtype_equal(a, b):
        raise ValueError("Input must be of same type")

    if a.ndim != 2 and b.ndim != 2:
        raise ValueError("Input must be 2-D.")

    if not (bhary.check(a) or bhary.check(b)):  # Both are regular numpy arrays
        return numpy.dot(a, b)
    else:
        a = array_create.array(a)
        b = array_create.array(b)

    # If the dtypes are both float, we can use BLAS to calculate
    # the dot-product, if BLAS is present.
    if not no_blas and a.dtype.kind in np.typecodes["AllFloat"] and b.dtype.kind in np.typecodes["AllFloat"]:
        try:
            return blas.gemm(a, b)
        except:
            pass

    return ufuncs.add.reduce(a[:, numpy.newaxis] * numpy.transpose(b), -1)


@fix_biclass_wrapper
def dot(a, b, no_blas=False):
    """
    Dot product of two arrays.

    For 2-D arrays it is equivalent to matrix multiplication, and for 1-D
    arrays to inner product of vectors (without complex conjugation). For
    N dimensions it is a sum product over the last axis of `a` and
    the second-to-last of `b`::

        dot(a, b)[i,j,k,m] = sum(a[i,j,:] * b[k,:,m])

    Parameters
    ----------
    a : array_like
        First argument.
    b : array_like
        Second argument.

    Returns
    -------
    output : ndarray
        Returns the dot product of `a` and `b`.  If `a` and `b` are both
        scalars or both 1-D arrays then a scalar is returned; otherwise
        an array is returned.

    Raises
    ------
    ValueError
        If the last dimension of `a` is not the same size as
        the second-to-last dimension of `b`.

    See Also
    --------
    vdot : Complex-conjugating dot product.
    tensordot : Sum products over arbitrary axes.
    einsum : Einstein summation convention.

    Examples
    --------
    >>> np.dot(3, 4)
    12

    Neither argument is complex-conjugated:

    >>> np.dot([2j, 3j], [2j, 3j])
    (-13+0j)

    For 2-D arrays it's the matrix product:

    >>> a = [[1, 0], [0, 1]]
    >>> b = [[4, 1], [2, 2]]
    >>> np.dot(a, b)
    array([[4, 1],
           [2, 2]])

    >>> a = np.arange(3*4*5*6).reshape((3,4,5,6))
    >>> b = np.arange(3*4*5*6)[::-1].reshape((5,4,6,3))
    >>> np.dot(a, b)[2,3,2,1,2,2]
    499128
    >>> sum(a[2,3,2,:] * b[1,2,:,2])
    499128

    """
    if not (bhary.check(a) or bhary.check(b)):  # Both are regular numpy arrays
        return numpy.dot(a, b)
    else:
        a = array_create.array(a)
        b = array_create.array(b)

    if b.ndim == 1:
        return ufuncs.add.reduce(a * b, -1)

    if a.ndim == 1:
        return ufuncs.add.reduce(a * numpy.transpose(b), -1)

    if not no_blas and a.ndim == 2 and b.ndim == 2:
        # If the dtypes are both float, we can use BLAS to calculate
        # the dot-product, if BLAS is present.
        if a.dtype.kind in np.typecodes["AllFloat"] and b.dtype.kind in np.typecodes["AllFloat"]:
            try:
                return blas.gemm(a, b)
            except:
                pass

    return ufuncs.add.reduce(a[:, numpy.newaxis] * numpy.transpose(b), -1)


@fix_biclass_wrapper
def norm(x, ord=None, axis=None):
    """
    This version of norm is not fully compliant with the NumPy version,
    it only supports computing 2-norm of a vector.
    """

    if ord != None:
        raise NotImplementedError("Unsupported value param ord=%s" % ord)
    if axis != None:
        raise NotImplementedError("Unsupported value of param ord=%s" % axis)

    r = np.sum(x * x)
    if issubclass(np.dtype(r.dtype).type, np.integer):
        r_f32 = np.empty(r.shape, dtype=np.float32)
        r_f32[:] = r
        r = r_f32
    return np.sqrt(r)


@fix_biclass_wrapper
def tensordot(a, b, axes=2):
    """
    Compute tensor dot product along specified axes for arrays >= 1-D.

    Given two tensors (arrays of dimension greater than or equal to one),
    `a` and `b`, and an array_like object containing two array_like
    objects, ``(a_axes, b_axes)``, sum the products of `a`'s and `b`'s
    elements (components) over the axes specified by ``a_axes`` and
    ``b_axes``. The third argument can be a single non-negative
    integer_like scalar, ``N``; if it is such, then the last ``N``
    dimensions of `a` and the first ``N`` dimensions of `b` are summed
    over.

    Parameters
    ----------
    a, b : array_like, len(shape) >= 1
        Tensors to "dot".
    axes : variable type
        * integer_like scalar
          Number of axes to sum over (applies to both arrays); or
        * (2,) array_like, both elements array_like of the same length
          List of axes to be summed over, first sequence applying to `a`,
          second to `b`.

    See Also
    --------
    dot, einsum

    Notes
    -----
    When there is more than one axis to sum over - and they are not the last
    (first) axes of `a` (`b`) - the argument `axes` should consist of
    two sequences of the same length, with the first axis to sum over given
    first in both sequences, the second axis second, and so forth.

    Examples
    --------
    A "traditional" example:

    >>> a = np.arange(60.).reshape(3,4,5)
    >>> b = np.arange(24.).reshape(4,3,2)
    >>> c = np.tensordot(a,b, axes=([1,0],[0,1]))
    >>> c.shape
    (5, 2)
    >>> c
    array([[ 4400.,  4730.],
           [ 4532.,  4874.],
           [ 4664.,  5018.],
           [ 4796.,  5162.],
           [ 4928.,  5306.]])
    >>> # A slower but equivalent way of computing the same...
    >>> d = np.zeros((5,2))
    >>> for i in range(5):
    ...   for j in range(2):
    ...     for k in range(3):
    ...       for n in range(4):
    ...         d[i,j] += a[k,n,i] * b[n,k,j]
    >>> c == d
    array([[ True,  True],
           [ True,  True],
           [ True,  True],
           [ True,  True],
           [ True,  True]], dtype=bool)

    An extended example taking advantage of the overloading of + and \\*:

    >>> a = np.array(range(1, 9))
    >>> a.shape = (2, 2, 2)
    >>> A = np.array(('a', 'b', 'c', 'd'), dtype=object)
    >>> A.shape = (2, 2)
    >>> a; A
    array([[[1, 2],
            [3, 4]],
           [[5, 6],
            [7, 8]]])
    array([[a, b],
           [c, d]], dtype=object)

    >>> np.tensordot(a, A) # third argument default is 2
    array([abbcccdddd, aaaaabbbbbbcccccccdddddddd], dtype=object)

    >>> np.tensordot(a, A, 1)
    array([[[acc, bdd],
            [aaacccc, bbbdddd]],
           [[aaaaacccccc, bbbbbdddddd],
            [aaaaaaacccccccc, bbbbbbbdddddddd]]], dtype=object)

    >>> np.tensordot(a, A, 0) # "Left for reader" (result too long to incl.)
    array([[[[[a, b],
              [c, d]],
              ...

    >>> np.tensordot(a, A, (0, 1))
    array([[[abbbbb, cddddd],
            [aabbbbbb, ccdddddd]],
           [[aaabbbbbbb, cccddddddd],
            [aaaabbbbbbbb, ccccdddddddd]]], dtype=object)

    >>> np.tensordot(a, A, (2, 1))
    array([[[abb, cdd],
            [aaabbbb, cccdddd]],
           [[aaaaabbbbbb, cccccdddddd],
            [aaaaaaabbbbbbbb, cccccccdddddddd]]], dtype=object)

    >>> np.tensordot(a, A, ((0, 1), (0, 1)))
    array([abbbcccccddddddd, aabbbbccccccdddddddd], dtype=object)

    >>> np.tensordot(a, A, ((2, 1), (1, 0)))
    array([acccbbdddd, aaaaacccccccbbbbbbdddddddd], dtype=object)

    """
    try:
        iter(axes)
    except:
        axes_a = list(range(-axes, 0))
        axes_b = list(range(0, axes))
    else:
        axes_a, axes_b = axes
    try:
        na = len(axes_a)
        axes_a = list(axes_a)
    except TypeError:
        axes_a = [axes_a]
        na = 1
    try:
        nb = len(axes_b)
        axes_b = list(axes_b)
    except TypeError:
        axes_b = [axes_b]
        nb = 1

    a, b = np.array(a), np.array(b)
    as_ = a.shape
    nda = len(a.shape)
    bs = b.shape
    ndb = len(b.shape)
    equal = True
    if (na != nb):
        equal = False
    else:
        for k in range(na):
            if as_[axes_a[k]] != bs[axes_b[k]]:
                equal = False
                break
            if axes_a[k] < 0:
                axes_a[k] += nda
            if axes_b[k] < 0:
                axes_b[k] += ndb
    if not equal:
        raise ValueError("shape-mismatch for sum")

    # Move the axes to sum over to the end of "a"
    # and to the front of "b"
    notin = [k for k in range(nda) if k not in axes_a]
    newaxes_a = notin + axes_a
    N2 = 1
    for axis in axes_a:
        N2 *= as_[axis]
    newshape_a = (-1, N2)
    olda = [as_[axis] for axis in notin]

    notin = [k for k in range(ndb) if k not in axes_b]
    newaxes_b = axes_b + notin
    N2 = 1
    for axis in axes_b:
        N2 *= bs[axis]
    newshape_b = (N2, -1)
    oldb = [bs[axis] for axis in notin]

    at = a.transpose(newaxes_a).reshape(newshape_a)
    bt = b.transpose(newaxes_b).reshape(newshape_b)
    res = dot(at, bt)
    return res.reshape(olda + oldb)


def _solve_tridiagonal_omp(a, b, c, rhs):
    from string import Template
    from textwrap import dedent

    KERNEL = Template(dedent('''
        #include <stddef.h>

        void execute(
            const ${DTYPE} *a,
            const ${DTYPE} *b,
            const ${DTYPE} *c,
            const ${DTYPE} *d,
            ${DTYPE} *solution
        ){
            const size_t m = ${SYS_DEPTH};
            const size_t total_size = ${SIZE};

            #pragma omp parallel for
            for(ptrdiff_t idx = 0; idx < total_size; idx += m) {
                ${DTYPE} cp[${SYS_DEPTH}];
                ${DTYPE} dp[${SYS_DEPTH}];
                cp[0] = c[idx] / b[idx];
                dp[0] = d[idx] / b[idx];
                for (ptrdiff_t j = 1; j < m; ++j) {
                    const ${DTYPE} norm_factor = b[idx+j] - a[idx+j] * cp[j-1];
                    cp[j] = c[idx+j] / norm_factor;
                    dp[j] = (d[idx+j] - a[idx+j] * dp[j-1]) / norm_factor;
                }
                solution[idx + m-1] = dp[m-1];
                for (ptrdiff_t j=m-2; j >= 0; --j) {
                    solution[idx + j] = dp[j] - cp[j] * solution[idx + j+1];
                }
            }
        }
    '''))
    
    kernel = KERNEL.substitute(
        DTYPE=user_kernel.dtype_to_c99(a.dtype),
        SYS_DEPTH=a.shape[-1],
        SIZE=a.size
    )

    res = array_create.empty_like(a)
    user_kernel.execute(
        kernel,
        [user_kernel.make_behaving(v) for v in (a, b, c, rhs, res)]
    )
    return res


def _solve_tridiagonal_ocl(a, b, c, rhs, local_work_size=32):
    from string import Template
    from textwrap import dedent

    KERNEL = Template(dedent('''
        #pragma OPENCL EXTENSION cl_khr_fp64 : enable

        kernel void execute(
            const global ${DTYPE} *a,
            const global ${DTYPE} *b,
            const global ${DTYPE} *c,
            const global ${DTYPE} *d,
            global ${DTYPE} *solution
        ){
            const size_t m = ${SYS_DEPTH};
            const size_t total_size = ${SIZE};
            const size_t idx = get_global_id(0) * m;

            if (idx >= total_size) {
                return;
            }

            private ${DTYPE} cp[${SYS_DEPTH}];
            private ${DTYPE} dp[${SYS_DEPTH}];
            cp[0] = c[idx] / b[idx];
            dp[0] = d[idx] / b[idx];
            for (ptrdiff_t j = 1; j < m; ++j) {
                const ${DTYPE} norm_factor = b[idx+j] - a[idx+j] * cp[j-1];
                cp[j] = c[idx+j] / norm_factor;
                dp[j] = (d[idx+j] - a[idx+j] * dp[j-1]) / norm_factor;
            }
            solution[idx + m-1] = dp[m-1];
            for (ptrdiff_t j=m-2; j >= 0; --j) {
                solution[idx + j] = dp[j] - cp[j] * solution[idx + j+1];
            }
        }
    '''))
    
    kernel = KERNEL.substitute(
        DTYPE=user_kernel.dtype_to_c99(a.dtype),
        SYS_DEPTH=a.shape[-1],
        SIZE=a.size
    )

    res = array_create.empty_like(a)

    # assemble work group size
    global_size = res.size // res.shape[-1]
    global_size = local_work_size * (global_size // local_work_size + 1)

    user_kernel.execute(
        kernel,
        [user_kernel.make_behaving(v) for v in (a, b, c, rhs, res)],
        tag="opencl",
        param={"global_work_size": global_size, "local_work_size": local_work_size}
    )
    return res


@fix_biclass_wrapper
def solve_tridiagonal(a, b, c, rhs, backend=None, **kwargs):
    """
    Solver for tridiagonal systems,

    ..math::

        A x = b

    based on the Thomas algorithm (not unconditionally stable).

    If the input arrays have more than one dimension, this function vectorizes over all
    but the last axis in parallel. All inputs must have equal shape.

    :param a: Lower diagonal elements. a[..., 0] is not used.
    :param b: Main diagonal elements.
    :param c: Upper diagonal elements. c[..., -1] is not used.
    :param rhs: Right-hand side vector.
    :param backend: Computational backend to use (openmp, opencl).
    :param kwargs: Additional argument to backend-specific solver. The opencl backend accepts
       an argument "local_work_size" specifying the local work group size.
    :returns: Solution of the tridiagonal system(s). Has the same shape as the input arrays.
    """
    if not (a.shape == b.shape == c.shape == rhs.shape):
        raise ValueError("All inputs must have equal shapes")

    if a.shape[-1] < 2:
        raise ValueError("Last axis must contain at least 2 elements")
    
    # Cast to common dtype
    common_dtype = numpy.find_common_type([v.dtype for v in (a, b, c, rhs)], [])

    if not numpy.issubdtype(common_dtype, numpy.floating):
        common_dtype = numpy.float64

    a, b, c, rhs = (v.astype(common_dtype, copy=False) for v in (a, b, c, rhs))

    if backend is None:
        import bohrium_api
        if bohrium_api.stack_info.is_opencl_in_stack():
            backend = "opencl"
        else:
            backend = "openmp"

    if backend == "openmp":
        return _solve_tridiagonal_omp(a, b, c, rhs, **kwargs)

    if backend == "opencl":
        return _solve_tridiagonal_ocl(a, b, c, rhs, **kwargs)

    raise ValueError("unknown backend '%s' (must be 'openmp' or 'opencl')" % backend)


@fix_biclass_wrapper
def cg(A, b, x=None, tol=1e-5, force_niter=None):
    """
    Conjugate Gradient (CG) solver

    Implemented as example MATLAB code from <https://en.wikipedia.org/wiki/Conjugate_gradient_method>
    """
    # If no guess is given, set an empty guess
    if x is None:
        x = array_create.zeros_like(b)

    r = b - dot(A, x)
    p = r.copy()
    r_squared = r * r
    rsold = np.sum(r_squared)

    tol_squared = tol * tol
    i = 0
    while np.max(r_squared) > tol_squared or force_niter is not None:
        Ap = dot(A, p)
        alpha = rsold / dot(p, Ap)
        x = x + alpha * p
        r = r - alpha * Ap
        r_squared = r * r
        rsnew = np.sum(r_squared)

        p = r + (rsnew / rsold) * p
        rsold = rsnew
        if force_niter is not None and i >= force_niter:
            break
        i += 1
    return x



"""
Disk IO
=======================
"""

import warnings
import numpy_force as numpy
from . import array_create
from .bhary import fix_biclass_wrapper


@fix_biclass_wrapper
def load(file, mmap_mode=None, allow_pickle=True, fix_imports=True, encoding='ASCII', bohrium=True):
    """
    Load arrays or pickled objects from ``.npy``, ``.npz`` or pickled files.

    Parameters
    ----------
    file : file-like object or string
        The file to read. File-like objects must support the
        ``seek()`` and ``read()`` methods. Pickled files require that the
        file-like object support the ``readline()`` method as well.
    mmap_mode : {None, 'r+', 'r', 'w+', 'c'}, optional
        If not None, then memory-map the file, using the given mode (see
        `numpy.memmap` for a detailed description of the modes).  A
        memory-mapped array is kept on disk. However, it can be accessed
        and sliced like any ndarray.  Memory mapping is especially useful
        for accessing small fragments of large files without reading the
        entire file into memory.
    allow_pickle : bool, optional
        Allow loading pickled object arrays stored in npy files. Reasons for
        disallowing pickles include security, as loading pickled data can
        execute arbitrary code. If pickles are disallowed, loading object
        arrays will fail.
        Default: True
    fix_imports : bool, optional
        Only useful when loading Python 2 generated pickled files on Python 3,
        which includes npy/npz files containing object arrays. If `fix_imports`
        is True, pickle will try to map the old Python 2 names to the new names
        used in Python 3.
    encoding : str, optional
        What encoding to use when reading Python 2 strings. Only useful when
        loading Python 2 generated pickled files on Python 3, which includes
        npy/npz files containing object arrays. Values other than 'latin1',
        'ASCII', and 'bytes' are not allowed, as they can corrupt numerical
        data. Default: 'ASCII'

    Returns
    -------
    result : array, tuple, dict, etc.
        Data stored in the file. For ``.npz`` files, the returned instance
        of NpzFile class must be closed to avoid leaking file descriptors.

    Raises
    ------
    IOError
        If the input file does not exist or cannot be read.
    ValueError
        The file contains an object array, but allow_pickle=False given.

    See Also
    --------
    save, savez, savez_compressed, loadtxt
    memmap : Create a memory-map to an array stored in a file on disk.

    Notes
    -----
    - If the file contains pickle data, then whatever object is stored
      in the pickle is returned.
    - If the file is a ``.npy`` file, then a single array is returned.
    - If the file is a ``.npz`` file, then a dictionary-like object is
      returned, containing ``{filename: array}`` key-value pairs, one for
      each file in the archive.
    - If the file is a ``.npz`` file, the returned value supports the
      context manager protocol in a similar fashion to the open function::

        with load('foo.npz') as data:
            a = data['a']

      The underlying file descriptor is closed when exiting the 'with'
      block.

    Examples
    --------
    Store data to disk, and load it again:

    >>> np.save('/tmp/123', np.array([[1, 2, 3], [4, 5, 6]]))
    >>> np.load('/tmp/123.npy')
    array([[1, 2, 3],
           [4, 5, 6]])

    Store compressed data to disk, and load it again:

    >>> a=np.array([[1, 2, 3], [4, 5, 6]])
    >>> b=np.array([1, 2])
    >>> np.savez('/tmp/123.npz', a=a, b=b)
    >>> data = np.load('/tmp/123.npz')
    >>> data['a']
    array([[1, 2, 3],
           [4, 5, 6]])
    >>> data['b']
    array([1, 2])
    >>> data.close()

    Mem-map the stored array, and then access the second row
    directly from disk:

    >>> X = np.load('/tmp/123.npy', mmap_mode='r')
    >>> X[1, :]
    memmap([4, 5, 6])

    """

    f = numpy.load(file, mmap_mode, allow_pickle, fix_imports, encoding)

    if mmap_mode is not None:
        warnings.warn("Bohrium does not support the mmap_mode argument in load()")
        return f
    else:
        return array_create.array(f, bohrium=bohrium)


@fix_biclass_wrapper
def save(file, arr, allow_pickle=True, fix_imports=True):
    """
    Save an array to a binary file in NumPy ``.npy`` format.

    Parameters
    ----------
    file : file or str
        File or filename to which the data is saved.  If file is a file-object,
        then the filename is unchanged.  If file is a string, a ``.npy``
        extension will be appended to the file name if it does not already
        have one.
    allow_pickle : bool, optional
        Allow saving object arrays using Python pickles. Reasons for disallowing
        pickles include security (loading pickled data can execute arbitrary
        code) and portability (pickled objects may not be loadable on different
        Python installations, for example if the stored objects require libraries
        that are not available, and not all pickled data is compatible between
        Python 2 and Python 3).
        Default: True
    fix_imports : bool, optional
        Only useful in forcing objects in object arrays on Python 3 to be
        pickled in a Python 2 compatible way. If `fix_imports` is True, pickle
        will try to map the new Python 3 names to the old module names used in
        Python 2, so that the pickle data stream is readable with Python 2.
    arr : array_like
        Array data to be saved.

    See Also
    --------
    savez : Save several arrays into a ``.npz`` archive
    savetxt, load

    Notes
    -----
    For a description of the ``.npy`` format, see the module docstring
    of `numpy.lib.format` or the Numpy Enhancement Proposal
    http://docs.scipy.org/doc/numpy/neps/npy-format.html

    Examples
    --------
    >>> from tempfile import TemporaryFile
    >>> outfile = TemporaryFile()

    >>> x = np.arange(10)
    >>> np.save(outfile, x)

    >>> outfile.seek(0) # Only needed here to simulate closing & reopening file
    >>> np.load(outfile)
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

    """

    return numpy.save(file, array_create.array(arr, bohrium=False), allow_pickle, fix_imports)


@fix_biclass_wrapper
def _savez(file, args, kwds, compress, allow_pickle=True, pickle_kwargs=None):
    ary_list = []
    for a in args:
        ary_list.append(array_create.array(a, bohrium=False))

    return numpy.save(file, array_create.array(arr, bohrium=False), allow_pickle, fix_imports)

    # Import is postponed to here since zipfile depends on gzip, an optional
    # component of the so-called standard library.
    import zipfile
    # Import deferred for startup time improvement
    import tempfile

    if isinstance(file, basestring):
        if not file.endswith('.npz'):
            file = file + '.npz'

    namedict = kwds
    for i, val in enumerate(args):
        key = 'arr_%d' % i
        if key in namedict.keys():
            raise ValueError(
                "Cannot use un-named variables and keyword %s" % key)
        namedict[key] = val

    if compress:
        compression = zipfile.ZIP_DEFLATED
    else:
        compression = zipfile.ZIP_STORED

    zipf = zipfile_factory(file, mode="w", compression=compression)

    # Stage arrays in a temporary file on disk, before writing to zip.
    fd, tmpfile = tempfile.mkstemp(suffix='-numpy.npy')
    os.close(fd)
    try:
        for key, val in namedict.items():
            fname = key + '.npy'
            fid = open(tmpfile, 'wb')
            try:
                format.write_array(fid, np.asanyarray(val),
                                   allow_pickle=allow_pickle,
                                   pickle_kwargs=pickle_kwargs)
                fid.close()
                fid = None
                zipf.write(tmpfile, arcname=fname)
            finally:
                if fid:
                    fid.close()
    finally:
        os.remove(tmpfile)

    zipf.close()


@fix_biclass_wrapper
def savez(file, *args, **kwds):
    """
    Save several arrays into a single file in uncompressed ``.npz`` format.

    If arguments are passed in with no keywords, the corresponding variable
    names, in the ``.npz`` file, are 'arr_0', 'arr_1', etc. If keyword
    arguments are given, the corresponding variable names, in the ``.npz``
    file will match the keyword names.

    Parameters
    ----------
    file : str or file
        Either the file name (string) or an open file (file-like object)
        where the data will be saved. If file is a string, the ``.npz``
        extension will be appended to the file name if it is not already there.
    args : Arguments, optional
        Arrays to save to the file. Since it is not possible for Python to
        know the names of the arrays outside `savez`, the arrays will be saved
        with names "arr_0", "arr_1", and so on. These arguments can be any
        expression.
    kwds : Keyword arguments, optional
        Arrays to save to the file. Arrays will be saved in the file with the
        keyword names.

    Returns
    -------
    None

    See Also
    --------
    save : Save a single array to a binary file in NumPy format.
    savetxt : Save an array to a file as plain text.
    savez_compressed : Save several arrays into a compressed ``.npz`` archive

    Notes
    -----
    The ``.npz`` file format is a zipped archive of files named after the
    variables they contain.  The archive is not compressed and each file
    in the archive contains one variable in ``.npy`` format. For a
    description of the ``.npy`` format, see `numpy.lib.format` or the
    Numpy Enhancement Proposal
    http://docs.scipy.org/doc/numpy/neps/npy-format.html

    When opening the saved ``.npz`` file with `load` a `NpzFile` object is
    returned. This is a dictionary-like object which can be queried for
    its list of arrays (with the ``.files`` attribute), and for the arrays
    themselves.

    Examples
    --------
    >>> from tempfile import TemporaryFile
    >>> outfile = TemporaryFile()
    >>> x = np.arange(10)
    >>> y = np.sin(x)

    Using `savez` with \\*args, the arrays are saved with default names.

    >>> np.savez(outfile, x, y)
    >>> outfile.seek(0) # Only needed here to simulate closing & reopening file
    >>> npzfile = np.load(outfile)
    >>> npzfile.files
    ['arr_1', 'arr_0']
    >>> npzfile['arr_0']
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

    Using `savez` with \\**kwds, the arrays are saved with the keyword names.

    >>> outfile = TemporaryFile()
    >>> np.savez(outfile, x=x, y=y)
    >>> outfile.seek(0)
    >>> npzfile = np.load(outfile)
    >>> npzfile.files
    ['y', 'x']
    >>> npzfile['x']
    array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

    """

    ary_list = []
    for a in args:
        ary_list.append(array_create.array(a, bohrium=False))
    return numpy.savez(file, *ary_list, **kwds)


@fix_biclass_wrapper
def savez_compressed(file, *args, **kwds):
    """
    Save several arrays into a single file in compressed ``.npz`` format.

    If keyword arguments are given, then filenames are taken from the keywords.
    If arguments are passed in with no keywords, then stored file names are
    arr_0, arr_1, etc.

    Parameters
    ----------
    file : str
        File name of ``.npz`` file.
    args : Arguments
        Function arguments.
    kwds : Keyword arguments
        Keywords.

    See Also
    --------
    numpy.savez : Save several arrays into an uncompressed ``.npz`` file format
    numpy.load : Load the files created by savez_compressed.

    """

    ary_list = []
    for a in args:
        ary_list.append(array_create.array(a, bohrium=False))
    return numpy.savez_compressed(file, *ary_list, **kwds)


@fix_biclass_wrapper
def loadtxt(fname, dtype=float, comments='#', delimiter=None,
            converters=None, skiprows=0, usecols=None, unpack=False,
            ndmin=0, bohrium=True):
    """
    Load data from a text file.

    Each row in the text file must have the same number of values.

    Parameters
    ----------
    fname : file or str
        File, filename, or generator to read.  If the filename extension is
        ``.gz`` or ``.bz2``, the file is first decompressed. Note that
        generators should return byte strings for Python 3k.
    dtype : data-type, optional
        Data-type of the resulting array; default: float.  If this is a
        structured data-type, the resulting array will be 1-dimensional, and
        each row will be interpreted as an element of the array.  In this
        case, the number of columns used must match the number of fields in
        the data-type.
    comments : str or sequence, optional
        The characters or list of characters used to indicate the start of a
        comment;
        default: '#'.
    delimiter : str, optional
        The string used to separate values.  By default, this is any
        whitespace.
    converters : dict, optional
        A dictionary mapping column number to a function that will convert
        that column to a float.  E.g., if column 0 is a date string:
        ``converters = {0: datestr2num}``.  Converters can also be used to
        provide a default value for missing data (but see also `genfromtxt`):
        ``converters = {3: lambda s: float(s.strip() or 0)}``.  Default: None.
    skiprows : int, optional
        Skip the first `skiprows` lines; default: 0.
    usecols : sequence, optional
        Which columns to read, with 0 being the first.  For example,
        ``usecols = (1,4,5)`` will extract the 2nd, 5th and 6th columns.
        The default, None, results in all columns being read.
    unpack : bool, optional
        If True, the returned array is transposed, so that arguments may be
        unpacked using ``x, y, z = loadtxt(...)``.  When used with a structured
        data-type, arrays are returned for each field.  Default is False.
    ndmin : int, optional
        The returned array will have at least `ndmin` dimensions.
        Otherwise mono-dimensional axes will be squeezed.
        Legal values: 0 (default), 1 or 2.

        .. versionadded:: 1.6.0

    Returns
    -------
    out : ndarray
        Data read from the text file.

    See Also
    --------
    load, fromstring, fromregex
    genfromtxt : Load data with missing values handled as specified.
    scipy.io.loadmat : reads MATLAB data files

    Notes
    -----
    This function aims to be a fast reader for simply formatted files.  The
    `genfromtxt` function provides more sophisticated handling of, e.g.,
    lines with missing values.

    .. versionadded:: 1.10.0

    The strings produced by the Python float.hex method can be used as
    input for floats.

    Examples
    --------
    >>> from io import StringIO   # StringIO behaves like a file object
    >>> c = StringIO("0 1\\n2 3")
    >>> np.loadtxt(c)
    array([[ 0.,  1.],
           [ 2.,  3.]])

    >>> d = StringIO("M 21 72\\nF 35 58")
    >>> np.loadtxt(d, dtype={'names': ('gender', 'age', 'weight'),
    ...                      'formats': ('S1', 'i4', 'f4')})
    array([('M', 21, 72.0), ('F', 35, 58.0)],
          dtype=[('gender', '|S1'), ('age', '<i4'), ('weight', '<f4')])

    >>> c = StringIO("1,0,2\\n3,0,4")
    >>> x, y = np.loadtxt(c, delimiter=',', usecols=(0, 2), unpack=True)
    >>> x
    array([ 1.,  3.])
    >>> y
    array([ 2.,  4.])

    """

    f = numpy.loadtxt(fname, dtype, comments, delimiter, converters, skiprows, usecols, unpack, ndmin)
    return array_create.array(f, bohrium=bohrium)


@fix_biclass_wrapper
def savetxt(fname, X, fmt='%.18e', delimiter=' ', newline='\n', header='',
            footer='', comments='# '):
    """
    Save an array to a text file.

    Parameters
    ----------
    fname : filename or file handle
        If the filename ends in ``.gz``, the file is automatically saved in
        compressed gzip format.  `loadtxt` understands gzipped files
        transparently.
    X : array_like
        Data to be saved to a text file.
    fmt : str or sequence of strs, optional
        A single format (%10.5f), a sequence of formats, or a
        multi-format string, e.g. 'Iteration %d -- %10.5f', in which
        case `delimiter` is ignored. For complex `X`, the legal options
        for `fmt` are:
            a) a single specifier, `fmt='%.4e'`, resulting in numbers formatted
                like `' (%s+%sj)' % (fmt, fmt)`
            b) a full string specifying every real and imaginary part, e.g.
                `' %.4e %+.4j %.4e %+.4j %.4e %+.4j'` for 3 columns
            c) a list of specifiers, one per column - in this case, the real
                and imaginary part must have separate specifiers,
                e.g. `['%.3e + %.3ej', '(%.15e%+.15ej)']` for 2 columns
    delimiter : str, optional
        String or character separating columns.
    newline : str, optional
        String or character separating lines.

        .. versionadded:: 1.5.0
    header : str, optional
        String that will be written at the beginning of the file.

        .. versionadded:: 1.7.0
    footer : str, optional
        String that will be written at the end of the file.

        .. versionadded:: 1.7.0
    comments : str, optional
        String that will be prepended to the ``header`` and ``footer`` strings,
        to mark them as comments. Default: '# ',  as expected by e.g.
        ``numpy.loadtxt``.

        .. versionadded:: 1.7.0


    See Also
    --------
    save : Save an array to a binary file in NumPy ``.npy`` format
    savez : Save several arrays into an uncompressed ``.npz`` archive
    savez_compressed : Save several arrays into a compressed ``.npz`` archive

    Notes
    -----
    Further explanation of the `fmt` parameter
    (``%[flag]width[.precision]specifier``):

    flags:
        ``-`` : left justify

        ``+`` : Forces to precede result with + or -.

        ``0`` : Left pad the number with zeros instead of space (see width).

    width:
        Minimum number of characters to be printed. The value is not truncated
        if it has more characters.

    precision:
        - For integer specifiers (eg. ``d,i,o,x``), the minimum number of
          digits.
        - For ``e, E`` and ``f`` specifiers, the number of digits to print
          after the decimal point.
        - For ``g`` and ``G``, the maximum number of significant digits.
        - For ``s``, the maximum number of characters.

    specifiers:
        ``c`` : character

        ``d`` or ``i`` : signed decimal integer

        ``e`` or ``E`` : scientific notation with ``e`` or ``E``.

        ``f`` : decimal floating point

        ``g,G`` : use the shorter of ``e,E`` or ``f``

        ``o`` : signed octal

        ``s`` : string of characters

        ``u`` : unsigned decimal integer

        ``x,X`` : unsigned hexadecimal integer

    This explanation of ``fmt`` is not complete, for an exhaustive
    specification see [1]_.

    References
    ----------
    .. [1] `Format Specification Mini-Language
           <http://docs.python.org/library/string.html#
           format-specification-mini-language>`_, Python Documentation.

    Examples
    --------
    >>> x = y = z = np.arange(0.0,5.0,1.0)
    >>> np.savetxt('test.out', x, delimiter=',')   # X is an array
    >>> np.savetxt('test.out', (x,y,z))   # x,y,z equal sized 1D arrays
    >>> np.savetxt('test.out', x, fmt='%1.4e')   # use exponential notation

    """

    return numpy.savetxt(fname, array_create.array(X, bohrium=False), fmt=fmt, delimiter=delimiter,
                         newline=newline, header=header, footer=footer, comments=comments)


@fix_biclass_wrapper
def fromregex(file, regexp, dtype, bohrium=True):
    """
    Construct an array from a text file, using regular expression parsing.

    The returned array is always a structured array, and is constructed from
    all matches of the regular expression in the file. Groups in the regular
    expression are converted to fields of the structured array.

    Parameters
    ----------
    file : str or file
        File name or file object to read.
    regexp : str or regexp
        Regular expression used to parse the file.
        Groups in the regular expression correspond to fields in the dtype.
    dtype : dtype or list of dtypes
        Dtype for the structured array.

    Returns
    -------
    output : ndarray
        The output array, containing the part of the content of `file` that
        was matched by `regexp`. `output` is always a structured array.

    Raises
    ------
    TypeError
        When `dtype` is not a valid dtype for a structured array.

    See Also
    --------
    fromstring, loadtxt

    Notes
    -----
    Dtypes for structured arrays can be specified in several forms, but all
    forms specify at least the data type and field name. For details see
    `doc.structured_arrays`.

    Examples
    --------
    >>> f = open('test.dat', 'w')
    >>> f.write("1312 foo\\n1534  bar\\n444   qux")
    >>> f.close()

    >>> regexp = r"(\\d+)\\s+(...)"  # match [digits, whitespace, anything]
    >>> output = np.fromregex('test.dat', regexp,
    ...                       [('num', np.int64), ('key', 'S3')])
    >>> output
    array([(1312L, 'foo'), (1534L, 'bar'), (444L, 'qux')],
          dtype=[('num', '<i8'), ('key', '|S3')])
    >>> output['num']
    array([1312, 1534,  444], dtype=int64)

    """

    return array_create.array(numpy.fromregex(file, regexp, dtype), bohrium=bohrium)


@fix_biclass_wrapper
def print_to_file(arr, fid, sep="", format="%s"):
    """
    a.tofile(arr, fid, sep="", format="%s")

    Write array 'arr' to a file as text or binary (default).

    Data is always written in 'C' order, independent of the order of `a`.
    The data produced by this method can be recovered using the function
    fromfile().

    Parameters
    ----------
    arr : array_like
        Array data to be saved.
    fid : file or str
        An open file object, or a string containing a filename.
    sep : str
        Separator between array items for text output.
        If "" (empty), a binary file is written, equivalent to
        ``file.write(a.tobytes())``.
    format : str
        Format string for text file output.
        Each entry in the array is formatted to text by first converting
        it to the closest Python type, and then using "format" % item.

    Notes
    -----
    This is a convenience function for quick storage of array data.
    Information on endianness and precision is lost, so this method is not a
    good choice for files intended to archive data or transport data between
    machines with different endianness. Some of these problems can be overcome
    by outputting the data as text files, at the expense of speed and file
    size.
    """

    f = array_create.array(arr, bohrium=False)
    return f.tofile(fid, sep=sep, format=format)



#!/usr/bin/env python
"""
This file is part of Bohrium and copyright (c) 2012 the Bohrium
http://bohrium.bitbucket.org

Bohrium is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

Bohrium is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the
GNU Lesser General Public License along with Bohrium.

If not, see <http://www.gnu.org/licenses/>.
"""
import collections
import numpy_force as np
from bohrium_api import _info

def dtype_of(obj):
    """Returns the dtype of 'obj'."""

    if isinstance(obj, np.dtype):
        return obj
    elif isinstance(obj, basestring):
        tmp = obj
    elif isinstance(obj, type):
        tmp = obj
    elif hasattr(obj, "dtype"):
        tmp = obj.dtype
    else:
        tmp = type(obj)
    return np.dtype(tmp)

def dtype_equal(*args):
    """Returns True when all 'args' is the same dtype."""

    if len(args) > 1:
        first_tmp = dtype_of(args[0])
        for tmp in args[1:]:
            if first_tmp is not dtype_of(tmp):
                return False
    return True

def dtype_in(dtype, dtypes):
    """Returns True when 'dtype' is in the list of 'dtypes'."""

    for datatype in dtypes:
        if dtype_equal(dtype, datatype):
            return True
    return False

def dtype_is_float(dtype):
    """Returns True when 'dtype' is a float or complex."""

    return dtype_in(dtype, [np.float32, np.float64, np.complex64, np.complex128])

def dtype_name(obj):
    """Returns the Bohrium name of the data type of the object 'obj'."""

    dtype = dtype_of(obj)
    if dtype_in(dtype, [np.bool_, np.bool, bool]):
        return 'bool8'
    else:
        return dtype.name

def type_sig(op_name, inputs):
    """
    Returns the type signature (output, input) to use with the given operation.
    NB: we only returns the type of the first input thus all input types must
        be identical
    """

    func = _info.op[op_name]
    #Note that we first use the dtype before the array as inputs to result_type()
    inputs = [getattr(t, 'dtype', t) for t in inputs]
    dtype = np.result_type(*inputs)
    for sig in func['type_sig']:
        if dtype.name == sig[1]:
            return (np.dtype(sig[0]), np.dtype(sig[1]))

    # Let's try use a float signature for the integer input
    if np.issubdtype(dtype, np.integer):
        for sig in func['type_sig']:
            if 'float' in sig[1]:
                return (np.dtype(sig[0]), np.dtype(sig[1]))

    raise TypeError("The ufunc bohrium.%s() does not support input data type: %s." % (op_name, dtype.name))

def dtype_support(dtype):
    """Returns True when Bohrium supports 'dtype' """

    if dtype_in(dtype, _info.numpy_types()):
        return True
    else:
        return False

def totalsize(array_like):
    """Return the total size of an list like object such as a bharray, ndarray, list, etc."""
    if np.isscalar(array_like):
        return 1
    elif hasattr(array_like, "size"):
        return array_like.size
    elif isinstance(array_like, collections.Iterable) and not isinstance(array_like, basestring):
        return sum(totalsize(item) for item in array_like)
    else:
        return 1

def is_scalar(a):
    """Is `a` a scalar type or 0-dim array?"""
    return np.isscalar(a) or a.ndim == 0


import bohrium as np
from sys import stderr
from . import ufuncs


def __lapack(name, a, b):
    if not b.flags['C_CONTIGUOUS']:
        b = b.copy(order='C')

    if not a.flags['C_CONTIGUOUS']:
        a = a.copy(order='C')

    ufuncs.extmethod(name, b, a, b)  # modifies 'b' unless copy is True
    return b


def gesv(a, b):
    return __lapack("lapack_gesv", a.T.copy(order='C'), b)


def gbsv(a, b):
    return __lapack("lapack_gbsv", a, b)


def gtsv(a, b):
    return __lapack("lapack_gtsv", a, b)


def posv(a, b):
    return __lapack("lapack_posv", a, b)


def ppsv(a, b):
    return __lapack("lapack_ppsv", a, b)


def spsv(a, b):
    return __lapack("lapack_spsv", a, b)



"""
Array Creation Routines
=======================
"""
import math
import warnings
import collections
from . import bhary
from bohrium_api import _info
from .bhary import fix_biclass_wrapper
import numpy_force as numpy
from ._util import dtype_equal, dtype_in, dtype_support


def _warn_dtype(dtype, stacklevel):
    """Raise a dtype-not-supported warning """
    warnings.warn("Bohrium does not support the dtype '%s', the new array will be a regular NumPy array."
                  % dtype, UserWarning, stacklevel)


# Notice, array() is not decorated with @fix_biclass_wrapper() since @fix_biclass_wrapper() calls bohrium.array(), which
# would result in an infinite recursion. Similarly, when calling numpy.array() we set the 'fix_biclass=False'
# argument, which prevent any further calls to bohrium.array().
def array(obj, dtype=None, copy=False, order=None, subok=False, ndmin=0, bohrium=True):
    """
    Create an array -- Bohrium or NumPy ndarray.

    Parameters
    ----------
    obj : array_like
        An array, any object exposing the array interface, an
        object whose __array__ method returns an array, or any
        (nested) sequence.
    dtype : data-type, optional
        The desired data-type for the array.  If not given, then
        the type will be determined as the minimum type required
        to hold the objects in the sequence.  This argument can only
        be used to 'upcast' the array.  For downcasting, use the
        .astype(t) method.
    copy : bool, optional
        If true, then the object is copied.  Otherwise, a copy
        will only be made if __array__ returns a copy, if obj is a
        nested sequence, or if a copy is needed to satisfy any of the other
        requirements (`dtype`, `order`, etc.).
    order : {'C', 'F', 'A'}, optional
        Specify the order of the array.  If order is 'C' (default), then the
        array will be in C-contiguous order (last-index varies the
        fastest).  If order is 'F', then the returned array
        will be in Fortran-contiguous order (first-index varies the
        fastest).  If order is 'A', then the returned array may
        be in any order (either C-, Fortran-contiguous, or even
        discontiguous).
    subok : bool, optional
        If True, then sub-classes will be passed-through, otherwise
        the returned array will be forced to be a base-class array (default).
    ndmin : int, optional
        Specifies the minimum number of dimensions that the resulting
        array should have.  Ones will be pre-pended to the shape as
        needed to meet this requirement.
    bohrium : boolean, optional
        Determines whether it is a Bohrium array (bohrium.ndarray) or a
        regular NumPy array (numpy.ndarray)

    Returns
    -------
    out : ndarray
        An array object satisfying the specified requirements.

    See Also
    --------
    empty, empty_like, zeros, zeros_like, ones, ones_like, fill

    Examples
    --------
    >>> np.array([1, 2, 3])
    array([1, 2, 3])

    Upcasting:

    >>> np.array([1, 2, 3.0])
    array([ 1.,  2.,  3.])

    More than one dimension:

    >>> np.array([[1, 2], [3, 4]])
    array([[1, 2],
           [3, 4]])

    Minimum dimensions 2:

    >>> np.array([1, 2, 3], ndmin=2)
    array([[1, 2, 3]])

    Type provided:

    >>> np.array([1, 2, 3], dtype=complex)
    array([ 1.+0.j,  2.+0.j,  3.+0.j])

    Data-type consisting of more than one element:

    >>> x = np.array([(1,2),(3,4)],dtype=[('a','<i4'),('b','<i4')])
    >>> x['a']
    array([1, 3])

    Creating an array from sub-classes:

    >>> np.array(np.mat('1 2; 3 4'))
    array([[1, 2],
           [3, 4]])

    >>> np.array(np.mat('1 2; 3 4'), subok=True)
    matrix([[1, 2],
            [3, 4]])

    """
    ary = obj
    if bohrium:
        if bhary.check(ary):
            if order == 'F':
                raise ValueError("Cannot convert a Bohrium array to column-major ('F') memory representation")
            elif order == 'C' and not ary.flags['C_CONTIGUOUS']:
                copy = True  # We need to copy in order to make the returned array contiguous

            if copy:
                t = empty_like(ary)
                t[...] = ary
                ary = t

            if dtype is not None and not dtype_equal(dtype, ary.dtype):
                t = empty_like(ary, dtype=dtype)
                t[...] = ary
                ary = t

            for i in range(ary.ndim, ndmin):
                ary = numpy.expand_dims(ary, i)

            return ary
        else:
            # Let's convert the array using regular NumPy.
            # When `ary` is not a regular NumPy array, we make sure that `ary` contains no Bohrium arrays
            if isinstance(ary, collections.Sequence) and \
                    not (isinstance(ary, numpy.ndarray) and ary.dtype.isbuiltin == 1):
                ary = list(ary)  # Let's make sure that `ary` is mutable
                for i in range(len(ary)):  # Converting 1-element Bohrium arrays to NumPy scalars
                    if bhary.check(ary[i]):
                        ary[i] = ary[i].copy2numpy()
            ary = numpy.array(ary, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin, fix_biclass=False)

            # In any case, the array must meet some requirements
            ary = numpy.require(ary, requirements=['C_CONTIGUOUS', 'ALIGNED', 'OWNDATA'])

            if bohrium and not dtype_support(ary.dtype):
                _warn_dtype(ary.dtype, 3)
                return ary

            ret = empty(ary.shape, dtype=ary.dtype)
            if ret.size > 0:
                ret._data_fill(ary)
            return ret
    else:
        if bhary.check(ary):
            ret = ary.copy2numpy()
            return numpy.array(ret, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin, fix_biclass=False)
        else:
            return numpy.array(ary, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin, fix_biclass=False)


@fix_biclass_wrapper
def empty(shape, dtype=float, bohrium=True):
    """
    Return a new matrix of given shape and type, without initializing entries.

    Parameters
    ----------
    shape : int or tuple of int
        Shape of the empty matrix.
    dtype : data-type, optional
        Desired output data-type.

    See Also
    --------
    empty_like, zeros

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    `empty`, unlike `zeros`, does not set the matrix values to zero,
    and may therefore be marginally faster.  On the other hand, it requires
    the user to manually set all the values in the array, and should be
    used with caution.

    Examples
    --------
    >>> import numpy.matlib
    >>> np.matlib.empty((2, 2))    # filled with random data
    matrix([[  6.76425276e-320,   9.79033856e-307],
            [  7.39337286e-309,   3.22135945e-309]])        #random
    >>> np.matlib.empty((2, 2), dtype=int)
    matrix([[ 6600475,        0],
            [ 6586976, 22740995]])                          #random

    """

    if bohrium and not dtype_support(dtype):
        _warn_dtype(dtype, 3)
        bohrium = False

    if not bohrium:
        return numpy.ndarray(shape, dtype=dtype)

    from . import _bh
    return _bh.ndarray(shape, dtype=dtype)


@fix_biclass_wrapper
def ones(shape, dtype=float, bohrium=True):
    """
    Matrix of ones.

    Return a matrix of given shape and type, filled with ones.

    Parameters
    ----------
    shape : {sequence of ints, int}
        Shape of the matrix
    dtype : data-type, optional
        The desired data-type for the matrix, default is np.float64.
    bohrium : boolean, optional
        Determines whether it is a Bohrium-enabled array or a regular NumPy array

    Returns
    -------
    out : matrix
        Matrix of ones of given shape, dtype, and order.

    See Also
    --------
    ones : Array of ones.
    matlib.zeros : Zero matrix.

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    If `shape` has length one i.e. ``(N,)``, or is a scalar ``N``,
    `out` becomes a single row matrix of shape ``(1,N)``.

    Examples
    --------
    >>> np.matlib.ones((2,3))
    matrix([[ 1.,  1.,  1.],
            [ 1.,  1.,  1.]])

    >>> np.matlib.ones(2)
    matrix([[ 1.,  1.]])

    """

    if bohrium and not dtype_support(dtype):
        _warn_dtype(dtype, 3)
        return numpy.ones(shape, dtype=dtype)

    A = empty(shape, dtype=dtype, bohrium=bohrium)
    A[...] = A.dtype.type(1)
    return A


@fix_biclass_wrapper
def zeros(shape, dtype=float, bohrium=True):
    """
    Return a matrix of given shape and type, filled with zeros.

    Parameters
    ----------
    shape : int or sequence of ints
        Shape of the matrix
    dtype : data-type, optional
        The desired data-type for the matrix, default is float.
    bohrium : boolean, optional
        Determines whether it is a Bohrium-enabled array or a regular NumPy array

    Returns
    -------
    out : matrix
        Zero matrix of given shape, dtype, and order.

    See Also
    --------
    numpy.zeros : Equivalent array function.
    matlib.ones : Return a matrix of ones.

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    If `shape` has length one i.e. ``(N,)``, or is a scalar ``N``,
    `out` becomes a single row matrix of shape ``(1,N)``.

    Examples
    --------
    >>> import numpy.matlib
    >>> np.matlib.zeros((2, 3))
    matrix([[ 0.,  0.,  0.],
            [ 0.,  0.,  0.]])

    >>> np.matlib.zeros(2)
    matrix([[ 0.,  0.]])

    """
    if bohrium and not dtype_support(dtype):
        _warn_dtype(dtype, 3)
        return numpy.zeros(shape, dtype=dtype)

    a = empty(shape, dtype=dtype, bohrium=bohrium)
    a[...] = a.dtype.type(0)
    return a


@fix_biclass_wrapper
def empty_like(a, dtype=None, bohrium=None):
    """
    Return a new array with the same shape and type as a given array.

    Parameters
    ----------
    a : array_like
        The shape and data-type of `a` define these same attributes of the
        returned array.
    dtype : data-type, optional
        Overrides the data type of the result.
    bohrium : boolean, optional
        Determines whether it is a Bohrium-enabled array or a regular NumPy array

    Returns
    -------
    out : ndarray
        Array of uninitialized (arbitrary) data with the same
        shape and type as `a`.

    See Also
    --------
    ones_like : Return an array of ones with shape and type of input.
    zeros_like : Return an array of zeros with shape and type of input.
    empty : Return a new uninitialized array.
    ones : Return a new array setting values to one.
    zeros : Return a new array setting values to zero.

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    This function does *not* initialize the returned array; to do that use
    `zeros_like` or `ones_like` instead.  It may be marginally faster than
    the functions that do set the array values.

    Examples
    --------
    >>> a = ([1,2,3], [4,5,6])                         # a is array-like
    >>> np.empty_like(a)
    array([[-1073741821, -1073741821,           3],    #random
           [          0,           0, -1073741821]])
    >>> a = np.array([[1., 2., 3.],[4.,5.,6.]])
    >>> np.empty_like(a)
    array([[ -2.00000715e+000,   1.48219694e-323,  -2.00000572e+000],#random
           [  4.38791518e-305,  -2.00000715e+000,   4.17269252e-309]])
    """
    if dtype is None:
        dtype = a.dtype

    if bohrium is None:
        bohrium = bhary.check(a)

    if bohrium and not dtype_support(dtype):
        _warn_dtype(dtype, 3)
        return numpy.empty_like(a, dtype=dtype, subok=False)

    return empty(a.shape, dtype, bohrium)


@fix_biclass_wrapper
def zeros_like(a, dtype=None, bohrium=None):
    """
    Return an array of zeros with the same shape and type as a given array.

    With default parameters, is equivalent to ``a.copy().fill(0)``.

    Parameters
    ----------
    a : array_like
        The shape and data-type of `a` define these same attributes of
        the returned array.
    dtype : data-type, optional
        Overrides the data type of the result.
    bohrium : boolean, optional
        Determines whether it is a Bohrium-enabled array or a regular NumPy array

    Returns
    -------
    out : ndarray
        Array of zeros with the same shape and type as `a`.

    See Also
    --------
    ones_like : Return an array of ones with shape and type of input.
    empty_like : Return an empty array with shape and type of input.
    zeros : Return a new array setting values to zero.
    ones : Return a new array setting values to one.
    empty : Return a new uninitialized array.

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    Examples
    --------
    >>> x = np.arange(6)
    >>> x = x.reshape((2, 3))
    >>> x
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> np.zeros_like(x)
    array([[0, 0, 0],
           [0, 0, 0]])

    >>> y = np.arange(3, dtype=np.float)
    >>> y
    array([ 0.,  1.,  2.])
    >>> np.zeros_like(y)
    array([ 0.,  0.,  0.])

    """
    if dtype is None:
        dtype = a.dtype

    if bohrium is None:
        bohrium = bhary.check(a)

    if bohrium and not dtype_support(dtype):
        _warn_dtype(dtype, 3)
        return numpy.zeros_like(a, dtype=dtype)

    b = empty_like(a, dtype=dtype, bohrium=bohrium)
    b[...] = b.dtype.type(0)

    return b


@fix_biclass_wrapper
def ones_like(a, dtype=None, bohrium=None):
    """
    Return an array of ones with the same shape and type as a given array.

    With default parameters, is equivalent to ``a.copy().fill(1)``.

    Parameters
    ----------
    a : array_like
        The shape and data-type of `a` define these same attributes of
        the returned array.
    dtype : data-type, optional
        Overrides the data type of the result.
    bohrium : boolean, optional
        Determines whether it is a Bohrium-enabled array or a regular NumPy array

    Returns
    -------
    out : ndarray
        Array of zeros with the same shape and type as `a`.

    See Also
    --------
    zeros_like : Return an array of zeros with shape and type of input.
    empty_like : Return an empty array with shape and type of input.
    zeros : Return a new array setting values to zero.
    ones : Return a new array setting values to one.
    empty : Return a new uninitialized array.

    Notes
    -----
    The order of the data in memory is always row-major (C-style).

    Examples
    --------
    >>> x = np.arange(6)
    >>> x = x.reshape((2, 3))
    >>> x
    array([[0, 1, 2],
           [3, 4, 5]])
    >>> np.ones_like(x)
    array([[1, 1, 1],
           [1, 1, 1]])

    >>> y = np.arange(3, dtype=np.float)
    >>> y
    array([ 0.,  1.,  2.])
    >>> np.ones_like(y)
    array([ 1.,  1.,  1.])

    """
    if dtype is None:
        dtype = a.dtype

    if bohrium is None:
        bohrium = bhary.check(a)

    if bohrium and not dtype_support(dtype):
        _warn_dtype(dtype, 3)
        return numpy.ones_like(a, dtype=dtype)

    b = empty_like(a, dtype=dtype, bohrium=bohrium)
    b[...] = b.dtype.type(1)

    return b


@fix_biclass_wrapper
def arange(start, stop=None, step=1, dtype=None, bohrium=True):
    """
    arange([start,] stop[, step,], dtype=None)

    Return evenly spaced values within a given interval.

    Values are generated within the half-open interval ``[start, stop)``
    (in other words, the interval including `start` but excluding `stop`).
    For integer arguments the function is equivalent to the Python built-in
    `range <http://docs.python.org/lib/built-in-funcs.html>`_ function,
    but returns a ndarray rather than a list.

    When using a non-integer step, such as 0.1, the results will often not
    be consistent.  It is better to use ``linspace`` for these cases.

    Parameters
    ----------
    start : number, optional
        Start of interval.  The interval includes this value.  The default
        start value is 0.
    stop : number
        End of interval.  The interval does not include this value, except
        in some cases where `step` is not an integer and floating point
        round-off affects the length of `out`.
    step : number, optional
        Spacing between values.  For any output `out`, this is the distance
        between two adjacent values, ``out[i+1] - out[i]``.  The default
        step size is 1.  If `step` is specified, `start` must also be given.
    dtype : dtype
        The type of the output array.  If `dtype` is not given, infer the data
        type from the other input arguments.

    Returns
    -------
    out : ndarray
        Array of evenly spaced values.

        For floating point arguments, the length of the result is
        ``ceil((stop - start)/step)``.  Because of floating point overflow,
        this rule may result in the last element of `out` being greater
        than `stop`.

    See Also
    --------
    linspace : Evenly spaced numbers with careful handling of endpoints.
    ogrid: Arrays of evenly spaced numbers in N-dimensions
    mgrid: Grid-shaped arrays of evenly spaced numbers in N-dimensions

    Examples
    --------
    >>> np.arange(3)
    array([0, 1, 2])
    >>> np.arange(3.0)
    array([ 0.,  1.,  2.])
    >>> np.arange(3,7)
    array([3, 4, 5, 6])
    >>> np.arange(3,7,2)
    array([3, 5])

    """
    if not bohrium:
        return numpy.arange(start, stop, step, dtype)

    if stop is None:
        stop = start
        start = type(stop)(0)

    try:
        integers = (int, long)
    except:
        integers = (int,)
    if not (isinstance(stop, integers) and isinstance(start, integers)):
        raise ValueError("arange(): start and stop must be integers")

    if step == 0:
        raise ValueError("arange(): step cannot be zero")

    # Let's make sure that 'step' is always positive
    swap_back = False
    if step < 0:
        step *= -1
        (start, stop) = (stop, start)
        swap_back = True

    if start >= stop:
        return array([], dtype=dtype, bohrium=bohrium)

    size = int(math.ceil((float(stop) - float(start)) / float(step)))
    if dtype is None:
        dtype = numpy.int64
    else:
        start = numpy.dtype(dtype).type(start)
        stop = numpy.dtype(dtype).type(stop)
        step = numpy.dtype(dtype).type(step)

    result = simply_range(size, dtype=dtype)
    if swap_back:
        step *= -1
        (start, stop) = (stop, start)

    if step != 1:
        result *= step

    if start != 0:
        result += start

    return result


@fix_biclass_wrapper
def simply_range(size, dtype=numpy.uint64):
    from . import _bh
    try:
        integers = (int, long)
    except:
        integers = (int,)

    if not isinstance(size, integers):
        raise ValueError("size must be an integer")

    if size < 1:
        raise ValueError("size must be greater than 0")

    if dtype_in(dtype, [numpy.int8,
                        numpy.int16,
                        numpy.int32,
                        numpy.uint8,
                        numpy.uint16,
                        numpy.uint32,
                        numpy.float16,
                        numpy.float32,
                        numpy.complex64]):
        A = empty((size,), dtype=numpy.uint32, bohrium=True)
    else:
        A = empty((size,), dtype=numpy.uint64, bohrium=True)

    _bh.ufunc(_info.op["range"]['id'], (A,))

    if not dtype_equal(dtype, A.dtype):
        B = empty_like(A, dtype=dtype)
        B[...] = A[...]
        return B
    else:
        return A


@fix_biclass_wrapper
def linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=float, bohrium=True):
    """
    Return evenly spaced numbers over a specified interval.

    Returns `num` evenly spaced samples, calculated over the
    interval [`start`, `stop` ].

    The endpoint of the interval can optionally be excluded.

    Parameters
    ----------
    start : scalar
        The starting value of the sequence.
    stop : scalar
        The end value of the sequence, unless `endpoint` is set to False.
        In that case, the sequence consists of all but the last of ``num + 1``
        evenly spaced samples, so that `stop` is excluded.  Note that the step
        size changes when `endpoint` is False.
    num : int, optional
        Number of samples to generate. Default is 50.
    endpoint : bool, optional
        If True, `stop` is the last sample. Otherwise, it is not included.
        Default is True.
    retstep : bool, optional
        If True, return (`samples`, `step`), where `step` is the spacing
        between samples.

    Returns
    -------
    samples : ndarray
        There are `num` equally spaced samples in the closed interval
        ``[start, stop]`` or the half-open interval ``[start, stop)``
        (depending on whether `endpoint` is True or False).
    step : float (only if `retstep` is True)
        Size of spacing between samples.


    See Also
    --------
    arange : Similiar to `linspace`, but uses a step size (instead of the
             number of samples).
    logspace : Samples uniformly distributed in log space.

    Examples
    --------
    >>> np.linspace(2.0, 3.0, num=5)
        array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ])
    >>> np.linspace(2.0, 3.0, num=5, endpoint=False)
        array([ 2. ,  2.2,  2.4,  2.6,  2.8])
    >>> np.linspace(2.0, 3.0, num=5, retstep=True)
        (array([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ]), 0.25)

    Graphical illustration:

    >>> import matplotlib.pyplot as plt
    >>> N = 8
    >>> y = np.zeros(N)
    >>> x1 = np.linspace(0, 10, N, endpoint=True)
    >>> x2 = np.linspace(0, 10, N, endpoint=False)
    >>> plt.plot(x1, y, 'o')
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.plot(x2, y + 0.5, 'o')
    [<matplotlib.lines.Line2D object at 0x...>]
    >>> plt.ylim([-0.5, 1])
    (-0.5, 1)
    >>> plt.show()

    """
    if not bohrium:
        # TODO: add copy=False to .astype()
        return numpy.linspace(start, stop, num=num, endpoint=endpoint, retstep=retstep).astype(dtype)

    num = int(num)
    if num <= 0:
        return array([], dtype=dtype)

    if endpoint:
        if num == 1:
            return array([numpy.dtype(dtype).type(start)])
        step = (stop - start) / float((num - 1))
    else:
        step = (stop - start) / float(num)

    y = arange(num, dtype=dtype)
    if step != 1: y *= step
    if start != 0: y += start

    if retstep:
        return y, step
    else:
        return y


@bhary.fix_biclass_wrapper
def copy(a, order='K'):
    """
    Return an array copy of the given object.

    Parameters
    ----------
    a : array_like
        Input data.
    order : {'C', 'F', 'A', 'K'}, optional
        Controls the memory layout of the copy. 'C' means C-order,
        'F' means F-order, 'A' means 'F' if `a` is Fortran contiguous,
        'C' otherwise. 'K' means match the layout of `a` as closely
        as possible. (Note that this function and :meth:ndarray.copy are very
        similar, but have different default values for their order=
        arguments.)

    Returns
    -------
    arr : ndarray
        Array interpretation of `a`.

    Notes
    -----
    This is equivalent to

    >>> np.array(a, copy=True)                              #doctest: +SKIP

    Examples
    --------
    Create an array x, with a reference y and a copy z:

    >>> x = np.array([1, 2, 3])
    >>> y = x
    >>> z = np.copy(x)

    Note that, when we modify x, y changes, but not z:

    >>> x[0] = 10
    >>> x[0] == y[0]
    True
    >>> x[0] == z[0]
    False

    """
    return array(a, order=order, copy=True)


def identity(n, dtype=float, bohrium=True):
    """
    Return the identity array.
    The identity array is a square array with ones on
    the main diagonal.

    Parameters
    ----------
    n : int
        Number of rows (and columns) in `n` x `n` output.
    dtype : data-type, optional
        Data-type of the output.  Defaults to ``float``.
    bohrium : boolean, optional
        Determines whether it is a Bohrium-enabled array or a regular NumPy array

    Returns
    -------
    out : ndarray
        `n` x `n` array with its main diagonal set to one,
        and all other elements 0.
    Examples
    --------
    >>> np.identity(3)
    array([[ 1.,  0.,  0.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.]])
    """
    from .array_manipulation import diagonal

    ret = zeros((n, n), dtype=dtype, bohrium=bohrium)
    diagonal(ret)[:] = 1
    return ret



"""
The module initialization of npbackend/bohrium imports and exposes all methods
required to become a drop-in replacement for numpy.
"""
import sys

if 'numpy_force' not in sys.modules:
    import numpy

    sys.modules['numpy_force'] = numpy
    del numpy

# We import all of NumPy and overwrite with the objects we implement our self
from numpy_force import *
from numpy_force import testing
import warnings
import bohrium_api
from .version import __version__, __version_info__
from .array_create import *
from .array_manipulation import *
from .reorganization import *
from .contexts import *
from .ufuncs import UFUNCS
from .masking import *
from .bhary import check, fix_biclass
from bohrium_api._info import numpy_types
from ._util import is_scalar
from . import linalg
from .linalg import matmul, dot, tensordot
from .summations import *
from .disk_io import *
from .concatenate import *
from .ufuncs import _handle__array_ufunc__
from . import contexts
from . import interop_pyopencl
from . import interop_pycuda
from . import interop_numpy
from . import backend_messaging
from . import loop
from . import user_kernel
from .loop import do_while
from .nobh import bincount
from .contexts import EnableBohrium as Enable, DisableBohrium as Disable
from ._bh import flush

# In NumPy `correlate` and `convolve` only handles 1D arrays whereas in SciPy they handles ND arrays.
# However, NumPy and SciPy's functionality differ! Thus, the ND version cannot replace NumPy's 1D version.
from .signal import correlate1d as correlate, convolve1d as convolve
from .signal import correlate as correlate_scipy, convolve as convolve_scipy
from numpy_force import dtype

asarray = array
asanyarray = array

if __version_info__ > bohrium_api.__version_info__:
    warnings.warn("The version of Bohrium API (%s) is newer than Bohrium (%s). Please upgrade "
                  "Bohrium (e.g. `pip install bohrium --upgrade`)" % (__version__, bohrium_api.__version__))


def replace_numpy(function):
    def wrapper(*args, **kwargs):
        with contexts.EnableBohrium():
            # Run your function/program
            result = function(*args, **kwargs)
        return result

    return wrapper


# Expose all ufuncs
for _name, _f in UFUNCS.items():
    if _name != "identity":  # NumPy has a function named `identity`, which we shouldn't overwrite
        exec ("%s = _f" % _name)

# Aliases
_aliases = [
    ('abs', 'absolute'),
    ('round', 'round_'),
    ('conjugate', 'conj'),
    # We handle all of NumPy's max and min the same!
    # Since reductions in OpenMP ignores NaN values we do the same always
    ('fmin', 'minimum'),
    ('fmax', 'maximum'),
    ('nanmin', 'minimum.reduce'),
    ('nanmax', 'maximum.reduce'),
    ('amin', 'minimum.reduce'),
    ('amax', 'maximum.reduce'),
]
for _f, _t in _aliases:
    exec ("%s = %s" % (_f, _t))
cumsum = add.accumulate
cumprod = multiply.accumulate

# Expose all data types
for _t in numpy_types():
    exec ("%s = numpy.%s" % (_t.__str__(), _t.__str__()))

# Type aliases
_type_aliases = [
    ('bool', 'bool'),
    ('int', 'int'),
    ('uint', 'numpy.uint64'),
    ('float', 'float'),
    ('complex', 'complex'),
    ('fmod', 'mod'),
    ('mod', 'remainder')
]

for _f, _t in _type_aliases:
    exec ("%s = %s" % (_f, _t))

# Note that the following modules needs ufuncs and dtypes
from . import random123 as random

# Some modules (e.g. scipy) accesses '__all__' directly
__all__ = [x for x in dir() if not x.startswith("_")]

# Python v2 need '{set,get}_printoptions()' here
if sys.version_info[0] < 3:
    def set_printoptions(*args, **kwargs):
        numpy.core.arrayprint.set_printoptions(*args, **kwargs)


    def get_printoptions(*args, **kwargs):
        return numpy.core.arrayprint.get_printoptions()

# Let's bohriumify the exposed API
from . import bohriumify

bohriumify.modules()



#!/usr/bin/env python

# In this module we implement an "as numpy" hack which makes it possible to
# utilize Bohrium using the command line argument "python -m bohrium.as_numpy"

import sys
import os
import argparse
import bohrium
import bohrium_api
from . import version


@bohrium.replace_numpy
def execfile_wrapper(path):
    """execfile() does not exist in Python 3"""

    # We need this ugly code in order to avoid wrapping the script execution in a try/except construct
    try:
        execfile
    except NameError:
        import runpy
        return runpy.run_path(path, init_globals={}, run_name="__main__")
    return execfile(path, {"__name__": "__main__", "__file__": path})


if len(sys.argv) <= 1:
    sys.stderr.write(
        'ERR: the "-m bohrium" does not support interactive mode. Use `-m bohrium --info` for Bohrium info\n')
    sys.exit(-1)

parser = argparse.ArgumentParser(add_help=False)
parser.add_argument(
    '--info',
    action="store_true",
    default=False,
    help='Print Runtime Info'
)
(args, argv) = parser.parse_known_args()

# If there are more arguments than parsed, we are running a regular script
# Else we are running one of the build-in Bohrium scripts
if len(argv) > 0:
    # Set the module search path to the dir of the script
    sys.argv.pop(0)
    if len(sys.argv) > 0:
        sys.path[0] = os.path.dirname(os.path.abspath(sys.argv[0]))
    else:
        sys.path[0] = ""
    execfile_wrapper(sys.argv[0])
else:
    if args.info:
        print("----\nBohrium version: %s" % version.__version__)
        print(bohrium_api.stack_info.pprint())

        cmd = "import bohrium as bh; import numpy as np; assert((bh.arange(10) == np.arange(10)).all())"
        sys.stdout.write('Sanity Check: "%s"' % cmd)
        try:
            exec (cmd)
            sys.stdout.write(' - success!\n')
        except AssertionError as e:
            sys.stdout.write('\n')
            sys.stderr.write("ERROR - the sanity checked failed!\n")



"""
Reorganization of Array Elements Routines
===========================
"""
import warnings
import numpy_force as numpy
from . import bhary
from bohrium_api import _info
from ._util import is_scalar
from .bhary import fix_biclass_wrapper
from . import array_create
from . import array_manipulation
from . import ufuncs
from . import numpy_backport


@fix_biclass_wrapper
def gather(ary, indexes):
    """
    gather(ary, indexes)

    Gather elements from 'ary' selected by 'indexes'.
    The values of 'indexes' are absolute indexed into a flatten 'ary'
    The shape of the returned array equals indexes.shape.

    Parameters
    ----------
    ary  : array_like
        The array to gather elements from.
    indexes : array_like, interpreted as integers
        Array or list of indexes that will be gather from 'array'

    Returns
    -------
    r : ndarray
        The gathered array freshly-allocated.
    """
    from . import _bh

    ary = array_manipulation.flatten(array_create.array(ary))

    # Convert a scalar index to a 1-element array
    if is_scalar(indexes):
        indexes = [indexes]
        
    indexes = array_create.array(indexes, dtype=numpy.uint64, bohrium=True)
    ret = array_create.empty(indexes.shape, dtype=ary.dtype, bohrium=True)
    if ary.size == 0 or indexes.size == 0:
        return array_create.array([])

    _bh.ufunc(_info.op['gather']['id'], (ret, ary, indexes))
    return ret


@fix_biclass_wrapper
def take(a, indices, axis=None, out=None, mode='raise'):
    """
    Take elements from an array along an axis.

    This function does the same thing as "fancy" indexing (indexing arrays
    using arrays); however, it can be easier to use if you need elements
    along a given axis.

    Parameters
    ----------
    a : array_like
        The source array.
    indices : array_like, interpreted as integers
        The indices of the values to extract.

        .. versionadded:: 1.8.0

        Also allow scalars for indices.
    axis : int, optional
        The axis over which to select values. By default, the flattened
        input array is used.
    out : ndarray, optional
        If provided, the result will be placed in this array. It should
        be of the appropriate shape and dtype.
    mode : {'raise', 'wrap', 'clip'}, optional
        Specifies how out-of-bounds indices will behave.

        * 'raise' -- raise an error (default)
        * 'wrap' -- wrap around
        * 'clip' -- clip to the range

        'clip' mode means that all indices that are too large are replaced
        by the index that addresses the last element along that axis. Note
        that this disables indexing with negative numbers.

    Returns
    -------
    subarray : ndarray
        The returned array has the same type as `a`.

    See Also
    --------
    compress : Take elements using a boolean mask
    ndarray.take : equivalent method

    Examples
    --------
    >>> a = [4, 3, 5, 7, 6, 8]
    >>> indices = [0, 1, 4]
    >>> np.take(a, indices)
    array([4, 3, 6])

    In this example if `a` is an ndarray, "fancy" indexing can be used.

    >>> a = np.array(a)
    >>> a[indices]
    array([4, 3, 6])

    If `indices` is not one dimensional, the output also has these dimensions.

    >>> np.take(a, [[0, 1], [2, 3]])
    array([[4, 3],
           [5, 7]])
    """

    if not bhary.check(a):
        indices = array_create.array(indices, bohrium=False)
        return numpy.take(a, indices, axis=axis, out=out, mode=mode)

    if mode != "raise":
        warnings.warn("Bohrium only supports the 'raise' mode not '%s', "
                      "it will be handled by the original NumPy." % mode, UserWarning, 2)
        a = array_create.array(a, bohrium=False)
        indices = array_create.array(indices, bohrium=False)
        return numpy.take(a, indices, axis=axis, out=out, mode=mode)

    if axis is not None and a.ndim > 1:
        warnings.warn("Bohrium does not support the 'axis' argument, "
                      "it will be handled by the original NumPy.", UserWarning, 2)
        a = array_create.array(a, bohrium=False)
        indices = array_create.array(indices, bohrium=False)
        return numpy.take(a, indices, axis=axis, out=out, mode=mode)

    ret = gather(a, indices)
    if out is not None:
        out[...] = ret
        return out
    else:
        return ret


@fix_biclass_wrapper
def take_using_index_tuple(a, index_tuple, out=None):
    """
    Take elements from the array 'a' specified by 'index_tuple'
    This function is very similar to take(), but takes a tuple of index arrays rather than a single index array
    
    Parameters
    ----------
    a : array_like
        The source array.
    index_tuple : tuple of array_like, interpreted as integers
        Each array in the tuple specified the indices of the values to extract for that axis. 
        The number of arrays in 'index_tuple' must equal the number of dimension in 'a'

    out : ndarray, optional
        If provided, the result will be placed in this array. It should
        be of the appropriate shape and dtype.


    Returns
    -------
    subarray : ndarray
        The returned array has the same type as `a`.

    """

    if not bhary.check(a):
        ret = a[index_tuple]
        if out is not None:
            out[...] = ret
            return out
        else:
            return ret

    assert len(index_tuple) == a.ndim

    if a.size == 0:
        return array_create.array([], dtype=a.dtype)

    if a.ndim == 1:
        return take(a, index_tuple[0], out=out)

    # Make sure that all index arrays are uint64 bohrium arrays
    index_list = []
    for index in index_tuple:
        index_list.append(array_create.array(index, dtype=numpy.uint64, bohrium=True))
        if index_list[-1].size == 0:
            return array_create.empty((0,), dtype=a.dtype)

    # And then broadcast them into the same shape
    index_list = array_manipulation.broadcast_arrays(*index_list)[0]

    # Let's find the absolute index
    abs_index = index_list[-1].copy()
    stride = a.shape[-1]
    for i in range(len(index_list) - 2, -1, -1):  # Iterate backwards from index_list[-2]
        abs_index += index_list[i] * stride
        stride *= a.shape[i]

    # take() support absolute indices
    ret = take(a, abs_index).reshape(index_list[0].shape)
    if out is not None:
        out[...] = ret
        return out
    else:
        return ret


@fix_biclass_wrapper
def scatter(ary, indexes, values):
    """
    scatter(ary, indexes, values)

    Scatter 'values' into 'ary' selected by 'indexes'.
    The values of 'indexes' are absolute indexed into a flatten 'ary'
    The shape of 'indexes' and 'value' must be equal.

    Parameters
    ----------
    ary  : array_like
        The target array to write the values to.
    indexes : array_like, interpreted as integers
        Array or list of indexes that will be written to in 'ary'
    values : array_like
        Values to write into 'ary"
    """
    from . import _bh

    indexes = array_manipulation.flatten(array_create.array(indexes, dtype=numpy.uint64), always_copy=False)
    values = array_manipulation.flatten(array_create.array(values, dtype=ary.dtype), always_copy=False)

    assert indexes.shape == values.shape
    if ary.size == 0 or indexes.size == 0:
        return

    # In order to ensure a contiguous array, we do the scatter on a flatten copy
    flat = array_manipulation.flatten(ary, always_copy=True)
    _bh.ufunc(_info.op['scatter']['id'], (flat, values, indexes))
    ary[...] = flat.reshape(ary.shape)


@fix_biclass_wrapper
def put(a, ind, v, mode='raise'):
    """
    Replaces specified elements of an array with given values.

    The indexing works on the flattened target array. `put` is roughly
    equivalent to:

    ::

        a.flat[ind] = v

    Parameters
    ----------
    a : ndarray
        Target array.
    ind : array_like
        Target indices, interpreted as integers.
    v : array_like
        Values to place in `a` at target indices. If `v` is shorter than
        `ind` it will be repeated as necessary.
    mode : {'raise', 'wrap', 'clip'}, optional
        Specifies how out-of-bounds indices will behave.

        * 'raise' -- raise an error (default)
        * 'wrap' -- wrap around
        * 'clip' -- clip to the range

        'clip' mode means that all indices that are too large are replaced
        by the index that addresses the last element along that axis. Note
        that this disables indexing with negative numbers.

    See Also
    --------
    putmask, place, take

    Examples
    --------
    >>> a = np.arange(5)
    >>> np.put(a, [0, 2], [-44, -55])
    >>> a
    array([-44,   1, -55,   3,   4])

    >>> a = np.arange(5)
    >>> np.put(a, 22, -5, mode='clip')
    >>> a
    array([ 0,  1,  2,  3, -5])

    """

    if ind.size == 0:
        return  # Nothing to insert!

    if not bhary.check(a):
        return numpy.put(a, ind.astype(numpy.int64), v, mode=mode)

    if mode != "raise":
        warnings.warn("Bohrium only supports the 'raise' mode not '%s', "
                      "it will be handled by the original NumPy." % mode, UserWarning, 2)
        return numpy.put(a, ind, v, mode=mode)

    indexes = array_manipulation.flatten(array_create.array(ind, dtype=numpy.uint64), always_copy=False)
    values = array_manipulation.flatten(array_create.array(v, dtype=a.dtype), always_copy=False)

    # Now let's make the shape of 'indexes' and 'values' match
    if indexes.size > values.size:
        if values.size == 1:
            # When 'values' is a scalar, we can broadcast it to match 'indexes'
            values = numpy_backport.as_strided(values, shape=indexes.shape, strides=(0,))
        else:  # else we repeat 'values' enough times to be larger than 'indexes'
            values = numpy_backport.as_strided(values,
                                               shape=(indexes.size // values.size + 2, values.size),
                                               strides=(0, values.itemsize))
            values = array_manipulation.flatten(values, always_copy=False)

    # When 'values' is too large, we simple cut the end off
    if values.size > indexes.size:
        values = values[0:indexes.size]

    # Now that 'indexes' and 'values' have the same shape, we can call 'scatter'
    scatter(a, indexes, values)


@fix_biclass_wrapper
def put_using_index_tuple(a, index_tuple, v):
    """
    Replaces specified elements of an array with given values.
    This function is very similar to put(), but takes a tuple of index arrays rather than a single index array.
    The indexing works like fancy indexing:
    
    ::

        a[index_tuple] = v

    Parameters
    ----------
    a : array_like
        The source array.
    index_tuple : tuple of array_like, interpreted as integers
        Each array in the tuple specified the indices of the values to extract for that axis. 
        The number of arrays in 'index_tuple' must equal the number of dimension in 'a'
    v : array_like
        Values to place in `a`.

    Returns
    -------
    subarray : ndarray
        The returned array has the same type as `a`.
    """

    if not bhary.check(a):
        a[index_tuple] = array_create.array(v, bohrium=False)
        return

    v = array_create.array(v, bohrium=True)
    assert len(index_tuple) == a.ndim

    if a.size == 0:
        return

    if a.ndim == 1:
        return put(a, index_tuple[0], v)

    # Make sure that all index arrays are uint64 bohrium arrays
    index_list = []
    for index in index_tuple:
        index_list.append(array_create.array(index, dtype=numpy.uint64, bohrium=True))
        if index_list[-1].size == 0:
            return array_create.empty((0,), dtype=a.dtype)

    # And then broadcast them into the same shape
    index_list = array_manipulation.broadcast_arrays(*index_list)[0]

    # Let's find the absolute index
    abs_index = index_list[-1].copy()
    stride = a.shape[-1]
    for i in range(len(index_list) - 2, -1, -1):  # Iterate backwards from index_list[-2]
        abs_index += index_list[i] * stride
        stride *= a.shape[i]

    # put() support absolute indices
    put(a, abs_index, v)


@fix_biclass_wrapper
def cond_scatter(ary, indexes, values, mask):
    """
    scatter(ary, indexes, values, mask)

    Scatter 'values' into 'ary' selected by 'indexes' where 'mask' is true.
    The values of 'indexes' are absolute indexed into a flatten 'ary'
    The shape of 'indexes', 'value', and 'mask' must be equal.


    Parameters
    ----------
    ary  : array_like
        The target array to write the values to.
    indexes : array_like, interpreted as integers
        Array or list of indexes that will be written to in 'ary'
    values : array_like
        Values to write into 'ary'
    mask : array_like, interpreted as booleans
        A mask that specifies which indexes and values to include and exclude
    """
    from . import _bh

    indexes = array_manipulation.flatten(array_create.array(indexes, dtype=numpy.uint64), always_copy=False)
    values = array_manipulation.flatten(array_create.array(values, dtype=ary.dtype), always_copy=False)
    mask = array_manipulation.flatten(array_create.array(mask, dtype=numpy.bool), always_copy=False)

    assert (indexes.shape == values.shape and values.shape == mask.shape)
    if ary.size == 0 or indexes.size == 0:
        return

    # In order to ensure a contiguous array, we do the scatter on a flatten copy
    flat = array_manipulation.flatten(ary, always_copy=True)
    _bh.ufunc(_info.op['cond_scatter']['id'], (flat, values, indexes, mask))
    ary[...] = flat.reshape(ary.shape)


@fix_biclass_wrapper
def pack(ary, mask):
    """
    pack(ary, mask)

    Packing the elements of 'ary' specified by 'mask' into new array that are contiguous
    The values of 'indexes' are absolute indexed into a flatten 'ary'
    The shape of 'mask' and 'ary' must be equal.


    Parameters
    ----------
    ary  : array_like, read flatten
        The array to read from.
    mask : array_like, interpreted as a flatten boolean array
        A mask that specifies which indexes of 'ary' to read
    """

    ary = array_manipulation.flatten(array_create.array(ary), always_copy=False)
    mask = array_manipulation.flatten(array_create.array(mask, dtype=numpy.bool), always_copy=False)
    assert (ary.shape == mask.shape)
    if ary.size == 0 or mask.size == 0:
        return

    true_indexes = ufuncs.add.accumulate(mask)
    true_count = int(true_indexes[-1])
    if true_count == 0:
        return array_create.empty((0,), dtype=ary.dtype)
    else:
        ret = array_create.empty((true_count,), dtype=ary.dtype)
        cond_scatter(ret, true_indexes - 1, ary, mask)
        return ret


@fix_biclass_wrapper
def flatnonzero(a):
    """
    Return indices that are non-zero in the flattened version of a.
    This is equivalent to a.ravel().nonzero()[0].
    Parameters
    ----------
    a : ndarray
        Input array.
    Returns
    -------
    res : ndarray
        Output array, containing the indices of the elements of `a.ravel()`
        that are non-zero.
    See Also
    --------
    nonzero : Return the indices of the non-zero elements of the input array.
    ravel : Return a 1-D array containing the elements of the input array.
    Examples
    --------
    >>> x = np.arange(-2, 3)
    >>> x
    array([-2, -1,  0,  1,  2])
    >>> np.flatnonzero(x)
    array([0, 1, 3, 4])
    Use the indices of the non-zero elements as an index array to extract
    these elements:
    >>> x.ravel()[np.flatnonzero(x)]
    array([-2, -1,  1,  2])
    """

    if a.dtype is not numpy.bool:
        mask = a != 0
    new_indexes = array_create.arange(a.size, dtype=numpy.uint64)
    return pack(new_indexes, mask)


@fix_biclass_wrapper
def nonzero(a):
    """
    Return the indices of the elements that are non-zero.
    Returns a tuple of arrays, one for each dimension of `a`,
    containing the indices of the non-zero elements in that
    dimension. The values in `a` are always tested and returned in
    row-major, C-style order. The corresponding non-zero
    values can be obtained with::
        a[nonzero(a)]
    To group the indices by element, rather than dimension, use::
        transpose(nonzero(a))
    The result of this is always a 2-D array, with a row for
    each non-zero element.
    Parameters
    ----------
    a : array_like
        Input array.
    Returns
    -------
    tuple_of_arrays : tuple
        Indices of elements that are non-zero.
    See Also
    --------
    flatnonzero :
        Return indices that are non-zero in the flattened version of the input
        array.
    ndarray.nonzero :
        Equivalent ndarray method.
    count_nonzero :
        Counts the number of non-zero elements in the input array.
    Examples
    --------
    >>> x = np.eye(3)
    >>> x
    array([[ 1.,  0.,  0.],
           [ 0.,  1.,  0.],
           [ 0.,  0.,  1.]])
    >>> np.nonzero(x)
    (array([0, 1, 2]), array([0, 1, 2]))
    >>> x[np.nonzero(x)]
    array([ 1.,  1.,  1.])
    >>> np.transpose(np.nonzero(x))
    array([[0, 0],
           [1, 1],
           [2, 2]])
    A common use for ``nonzero`` is to find the indices of an array, where
    a condition is True.  Given an array `a`, the condition `a` > 3 is a
    boolean array and since False is interpreted as 0, np.nonzero(a > 3)
    yields the indices of the `a` where the condition is true.
    >>> a = np.array([[1,2,3],[4,5,6],[7,8,9]])
    >>> a > 3
    array([[False, False, False],
           [ True,  True,  True],
           [ True,  True,  True]], dtype=bool)
    >>> np.nonzero(a > 3)
    (array([1, 1, 1, 2, 2, 2]), array([0, 1, 2, 0, 1, 2]))
    The ``nonzero`` method of the boolean array can also be called.
    >>> (a > 3).nonzero()
    (array([1, 1, 1, 2, 2, 2]), array([0, 1, 2, 0, 1, 2]))
    """

    if a.ndim == 1:
        return (flatnonzero(a),)

    if not a.flags['C_CONTIGUOUS']:
        a = a.copy(order='C')

    nz = flatnonzero(a)
    ret = []
    for stride_in_bytes in a.strides:
        stride = stride_in_bytes // a.itemsize
        assert stride_in_bytes % a.itemsize == 0
        tmp = nz // stride
        ret.append(tmp)
        nz -= tmp * stride
    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
/*
This file is part of Bohrium and copyright (c) 2018 the Bohrium
<http://www.bh107.org>

Bohrium is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3
of the License, or (at your option) any later version.

Bohrium is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the
GNU Lesser General Public License along with Bohrium.

If not, see <http://www.gnu.org/licenses/>.
*/
"""

from setuptools import setup, find_packages
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext as setup_build_ext
from setuptools.command.build_py import build_py as setup_build_py
from setuptools.command.sdist import sdist as setup_sdist
import numbers
import os
import glob
import re

""" Beside the regular setup arguments, this script reads the follow environment variables:

      * USE_CYTHON - if defined, this setup will cythonize all pyx files.
    
    NB: when running the source distribution command, `setup.py sdist`, do it from the same directory as `setup.py` 
"""


def script_path(*paths):
    prefix = os.path.abspath(os.path.dirname(__file__))
    assert len(prefix) > 0
    return os.path.join(prefix, *paths)


def version_file_exist():
    """Return whether the version.py file exist or not"""
    ver_path = script_path("bohrium", "version.py")
    return os.path.exists(ver_path)


def get_version():
    """Returns the version and version_info.
        If the version.py file doesn't exist, the version of Bohrium API is returned.
        NB: If the version.py file doesn't exist, this function must be called after the call to `setup()`.
    """
    ver_path = script_path("bohrium", "version.py")
    if os.path.exists(ver_path):
        print("Getting version from version.py")
        # Loading `__version__` variable from the version file
        with open(ver_path, "r") as f:
            t=f.read()
            version = re.search("__version__\s*=\s*\"([^\"]*)\"",t).group(1);
            version_info = eval(re.search("__version_info__\s*=\s*(\([^\)]+\))",t).group(1));
            return (version, version_info)
    else:
        print("Getting version from bohrium_api")
        import bohrium_api
        return (bohrium_api.__version__, bohrium_api.__version_info__)


def get_bohrium_api_required_string():
    """Returns the install_requires/setup_requires string for `bohrium_api`"""
    try:
        ver_tuple = get_version()[1]
        return "bohrium_api>=%d.%d.%d" % (ver_tuple[0], ver_tuple[1], ver_tuple[2])
    except ImportError:
        return "bohrium_api"  # If `bohrium_api` is not available, we expect PIP to install the newest package


def get_pyx_extensions():
    """Find and compiles all cython extensions"""
    include_dirs = []
    if 'USE_CYTHON' in os.environ:
        import numpy
        import bohrium_api
        include_dirs.extend([numpy.get_include(), bohrium_api.get_include()])
    pyx_list = glob.glob(script_path("bohrium", "*.pyx"))
    ret = []
    for pyx in pyx_list:
        ret.append(Extension(name="bohrium.%s" % os.path.splitext(os.path.basename(pyx))[0],
                             sources=[pyx],
                             include_dirs=include_dirs))
    ret.append(Extension(name="bohrium.nobh.bincount_cython",
                         sources=[script_path("bohrium", "nobh", "bincount_cython.pyx")],
                         include_dirs=include_dirs))
    if 'USE_CYTHON' in os.environ:
        import Cython.Build
        return Cython.Build.cythonize(ret, nthreads=2)
    else:
        return ret


def gen_version_file_in_cmd(self, target_dir):
    """We extend the setup commands to also generate the `version.py` file if it doesn't exist already"""
    if not self.dry_run:
        version, version_info = get_version()
        if not version_file_exist():
            self.mkpath(target_dir)
            p = os.path.join(target_dir, 'version.py')
            print("Generating '%s'" % p)
            with open(p, 'w') as fobj:
                fobj.write("__version__ = \"%s\"\n" % version)
                fobj.write("__version_info__ = %s\n" % str(version_info))


class BuildPy(setup_build_py):
    def run(self):
        gen_version_file_in_cmd(self, os.path.join(self.build_lib, 'bohrium'))
        setup_build_py.run(self)


class Sdist(setup_sdist):
    def run(self):
        gen_version_file_in_cmd(self, "bohrium")
        setup_sdist.run(self)


class BuildExt(setup_build_ext):
    """We delay the numpy and bohrium dependency to the build command.
       Hopefully, PIP has installed them at this point."""

    def run(self):
        if not self.dry_run:
            import numpy
            import bohrium_api
            for ext in self.extensions:
                ext.include_dirs.extend([numpy.get_include(), bohrium_api.get_include()])
        setup_build_ext.run(self)


class DelayedVersion(numbers.Number):
    """In order to delay the version evaluation that depend on `bohrium_api`, we use this class"""

    def __str__(self):
        return get_version()[0]


setup(
    cmdclass={'build_ext': BuildExt, 'build_py': BuildPy, 'sdist': Sdist},
    name='bohrium',
    version=DelayedVersion(),
    description='Bohrium Python/NumPy Backend',
    long_description='Bohrium for Python <www.bh107.org>',
    url='http://bh107.org',
    author='The Bohrium Team',
    author_email='contact@bh107.org',
    maintainer='Mads R. B. Kristensen',
    maintainer_email='madsbk@gmail.com',
    platforms=['Linux', 'OSX'],

    # Choose your license
    license='GPL',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',

        'Topic :: Scientific/Engineering',
        'Topic :: Software Development',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License (GPL)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: C',
    ],

    # What does your project relate to?
    keywords='Bohrium, bh107, Python, C, HPC, MPI, PGAS, CUDA, OpenCL, OpenMP',

    # Dependencies
    install_requires=['numpy>=1.7', get_bohrium_api_required_string()],
    setup_requires=['numpy>=1.7', get_bohrium_api_required_string()],

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['tests']),

    ext_modules=get_pyx_extensions() + [
        Extension(
            name='bohrium._bh',
            sources=[script_path('src', '_bh.c'),
                     script_path('src', 'bharray.c'),
                     script_path('src', 'handle_array_op.c'),
                     script_path('src', 'handle_special_op.c'),
                     script_path('src', 'memory.c'),
                     script_path('src', 'util.c')],
            depends=[
                script_path('src', '_bh.h'),
                script_path('src', 'bharray.h'),
                script_path('src', 'util.h'),
                script_path('src', 'handle_array_op.h'),
                script_path('src', 'handle_special_op.h'),
                script_path('src', 'memory.h'),
                script_path('src', 'operator_overload.c')
            ],
            libraries=['dl'],
            extra_compile_args=["-std=c99"],
        )]
)



# This module duplicates the NumPy API

from numpy_force.dual import *
from numpy_force.dual import register_func, restore_func, restore_all

__all__ = ['fft', 'ifft', 'fftn', 'ifftn', 'fft2', 'ifft2',
           'norm', 'inv', 'svd', 'solve', 'det', 'eig', 'eigvals',
           'eigh', 'eigvalsh', 'lstsq', 'pinv', 'cholesky', 'i0']



"""
Send and receive a message through the Bohrium component stack
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

from ._bh import message as _backend_msg


def statistic_enable_and_reset():
    """Reset and enable the Bohrium statistic"""
    return _backend_msg("statistic_enable_and_reset")


def statistic():
    """Return a YAML string of Bohrium statistic"""
    return _backend_msg("statistic")


def gpu_disable():
    """Disable the GPU backend in the current runtime stack"""
    return _backend_msg("GPU: disable")


def gpu_enable():
    """Enable the GPU backend in the current runtime stack"""
    return _backend_msg("GPU: enable")


def runtime_info():
    """Return a YAML string describing the current Bohrium runtime"""
    return _backend_msg("info")


def cuda_use_current_context():
    """Tell the CUDA backend to use the current CUDA context (useful for PyCUDA interop)"""
    return _backend_msg("CUDA: use current context")



"""
NumPy backport implementations
==============================

We use some NumPy functions introduced in later version of NumPy.
This module implement those versions
"""

import numpy_force as numpy


# NB: The older version of NumPy `as_strided()` does'nt support `subok` so we include the implementation here and
#     makes `subok` default true.
def as_strided(x, shape=None, strides=None, subok=True, writeable=True):
    """
    Create a view into the array with the given shape and strides.
    .. warning:: This function has to be used with extreme care, see notes.
    Parameters
    ----------
    x : ndarray
        Array to create a new.
    shape : sequence of int, optional
        The shape of the new array. Defaults to ``x.shape``.
    strides : sequence of int, optional
        The strides of the new array. Defaults to ``x.strides``.
    subok : bool, optional
        .. versionadded:: 1.10
        If True, subclasses are preserved.
    writeable : bool, optional
        .. versionadded:: 1.12
        If set to False, the returned array will always be readonly.
        Otherwise it will be writable if the original array was. It
        is advisable to set this to False if possible (see Notes).
    Returns
    -------
    view : ndarray
    See also
    --------
    broadcast_to: broadcast an array to a given shape.
    reshape : reshape an array.
    Notes
    -----
    ``as_strided`` creates a view into the array given the exact strides
    and shape. This means it manipulates the internal data structure of
    ndarray and, if done incorrectly, the array elements can point to
    invalid memory and can corrupt results or crash your program.
    It is advisable to always use the original ``x.strides`` when
    calculating new strides to avoid reliance on a contiguous memory
    layout.
    Furthermore, arrays created with this function often contain self
    overlapping memory, so that two elements are identical.
    Vectorized write operations on such arrays will typically be
    unpredictable. They may even give different results for small, large,
    or transposed arrays.
    Since writing to these arrays has to be tested and done with great
    care, you may want to use ``writeable=False`` to avoid accidental write
    operations.
    For these reasons it is advisable to avoid ``as_strided`` when
    possible.
    """

    class DummyArray(object):
        """Dummy object that just exists to hang __array_interface__ dictionaries
        and possibly keep alive a reference to a base array.
        """

        def __init__(self, interface, base=None):
            self.__array_interface__ = interface
            self.base = base

    def _maybe_view_as_subclass(original_array, new_array):
        if type(original_array) is not type(new_array):
            # if input was an ndarray subclass and subclasses were OK,
            # then view the result as that subclass.
            new_array = new_array.view(type=type(original_array))
            # Since we have done something akin to a view from original_array, we
            # should let the subclass finalize (if it has it implemented, i.e., is
            # not None).
            if new_array.__array_finalize__:
                new_array.__array_finalize__(original_array)
        return new_array

    # first convert input to array, possibly keeping subclass
    x = numpy.array(x, copy=False, subok=subok)
    interface = dict(x.__array_interface__)
    if shape is not None:
        interface['shape'] = tuple(shape)
    if strides is not None:
        interface['strides'] = tuple(strides)

    array = numpy.asarray(DummyArray(interface, base=x))
    # The route via `__interface__` does not preserve structured
    # dtypes. Since dtype should remain unchanged, we set it explicitly.
    array.dtype = x.dtype
    view = _maybe_view_as_subclass(x, array)

    if view.flags.writeable and not writeable:
        view.flags.writeable = False
    return view
